#!/usr/bin/env python3

import os
import tempfile
from pathlib import Path
from shutil import rmtree
from subprocess import run
from unittest import TestCase, main

from readalongs.align import calculate_adjustment, correct_adjustments
from readalongs.audio_utils import (
    join_section,
    mute_section,
    read_audio_from_file,
    remove_section,
    write_audio_to_file,
)
from readalongs.log import LOGGER


class TestAudio(TestCase):
    def setUp(self):
        self.data_path = os.path.join(os.path.dirname(__file__), "data")
        self.audio_segment = read_audio_from_file(
            os.path.join(self.data_path, "audio_sample.ogg")
        )
        self.noisy_segment = read_audio_from_file(
            os.path.join(self.data_path, "noise_at_1500.mp3")
        )
        # Use a TemporaryDirectory object for temp outputs, so they get cleaned
        # automatically cleaned even in case of errors or aborted runs.
        self.tempdirobj = tempfile.TemporaryDirectory(
            prefix="test_audio_tmpdir", dir="."
        )
        self.tempdir = self.tempdirobj.name

    def tearDown(self):
        self.tempdirobj.cleanup()

    def align(self, input_text_path, input_audio_path, output_path, flags):
        args = [
            "readalongs",
            "align",
            input_text_path,
            input_audio_path,
            output_path,
        ] + flags
        LOGGER.info(
            f"Aligning {input_text_path} and {input_audio_path}, outputting to {output_path}"
        )
        return run(args, capture_output=True)

    def test_mute_section(self):
        """ Should mute section of audio
        """
        muted_segment = mute_section(self.audio_segment, 1000, 2000)
        muted_section = muted_segment[1000:2000]
        self.assertLessEqual(muted_section.max, 1)

    def test_remove_section(self):
        """ Should remove section of audio
        """
        removed_segment = remove_section(self.audio_segment, 1000, 2000)
        self.assertNotEqual(len(removed_segment), len(self.audio_segment))
        self.assertEqual(len(removed_segment), len(self.audio_segment) - 1000)

    def test_rejoin_section(self):
        """ Should rejoin removed/muted sections
        """
        removed_section = self.audio_segment[1000:2000]
        removed_segment = remove_section(self.audio_segment, 1000, 2000)
        rejoined_segment = join_section(removed_segment, removed_section, 1000)
        self.assertEqual(len(rejoined_segment), len(self.audio_segment))

    def test_align_sample(self):
        """ Sanity check that test audio should align
        """
        # Align
        input_text_path = os.path.join(self.data_path, "audio_sample.txt")
        input_audio_path = os.path.join(self.data_path, "audio_sample.ogg")
        flags = ["-i", "-l", "eng"]
        output_path = os.path.join(self.tempdir, "output")
        log = self.align(input_text_path, input_audio_path, output_path, flags)
        LOGGER.info(str(log))
        # Check Result
        smilpath = Path(output_path)
        smil_files = smilpath.glob("*.smil")
        self.assertTrue(next(smil_files, False), "No *.smil files found")
        self.assertFalse("error" in str(log).lower())

    def test_align_removed(self):
        """ Try aligning section with removed audio
        """
        # Process Audio
        removed_segment = remove_section(self.noisy_segment, 1500, 2500)
        audio_output_path = os.path.join(self.tempdir, "removed_sample.mp3")
        removed_segment.export(audio_output_path)
        # Align
        input_text_path = os.path.join(self.data_path, "audio_sample.txt")
        input_audio_path = audio_output_path
        flags = ["-i", "-l", "eng"]
        output_path = os.path.join(self.tempdir, "output_removed")
        log = self.align(input_text_path, input_audio_path, output_path, flags)
        LOGGER.info(str(log))
        # Check Result
        smilpath = Path(output_path)
        smil_files = smilpath.glob("*.smil")
        self.assertTrue(next(smil_files, False), "No *.smil files found")
        self.assertFalse("error" in str(log).lower())

    def test_align_muted(self):
        """ Try aligning section with muted audio
        """
        # Process Audio
        muted_segment = mute_section(self.noisy_segment, 1500, 2500)
        audio_output_path = os.path.join(self.tempdir, "muted_sample.mp3")
        muted_segment.export(audio_output_path)
        # Align
        input_text_path = os.path.join(self.data_path, "audio_sample.txt")
        input_audio_path = audio_output_path
        flags = ["-i", "-l", "eng", "-b"]
        output_path = os.path.join(self.tempdir, "output_muted")
        log = self.align(input_text_path, input_audio_path, output_path, flags)
        LOGGER.info(str(log))
        # Check Result
        smilpath = Path(output_path)
        smil_files = smilpath.glob("*.smil")
        self.assertTrue(next(smil_files, False), "No *.smil files found")
        self.assertFalse("error" in str(log).lower())

    def test_adjustment_calculation(self):
        """ Try adjusting alignments of re-built audio
        """
        self.assertEqual(
            calculate_adjustment(1000, [{"begin": 1000, "end": 1100}]), 100
        )
        self.assertEqual(calculate_adjustment(900, [{"begin": 1000, "end": 1100}]), 0)
        self.assertEqual(
            calculate_adjustment(1000, [{"begin": 1000, "end": 1100}]), 100
        )
        self.assertEqual(calculate_adjustment(900, [{"begin": 1000, "end": 1100}]), 0)
        self.assertEqual(
            calculate_adjustment(2000, [{"begin": 1000, "end": 1100}]), 100
        )
        # Function only accepts args in ms
        self.assertNotEqual(
            calculate_adjustment(
                1.0, [{"begin": 1000, "end": 1100}, {"begin": 1500, "end": 1700}]
            ),
            100,
        )
        self.assertNotEqual(
            calculate_adjustment(
                2.0, [{"begin": 1000, "end": 1100}, {"begin": 1500, "end": 2100}]
            ),
            700,
        )

    def test_adjustment_correction(self):
        """ Try correcting adjusted alignments of re-built audio
        """
        self.assertEqual(
            correct_adjustments(950, 1125, [{"begin": 1000, "end": 1100}]), (950, 1000)
        )
        self.assertEqual(
            correct_adjustments(975, 1150, [{"begin": 1000, "end": 1100}]), (1100, 1150)
        )
        # Function only accepts args in ms
        self.assertNotEqual(
            correct_adjustments(0.950, 1.125, [{"begin": 1000, "end": 1100}]),
            (950, 1000),
        )
        self.assertNotEqual(
            correct_adjustments(0.975, 1.150, [{"begin": 1000, "end": 1100}]),
            (1100, 1150),
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import io
import os
import re
import sys
import tempfile
from shutil import copyfile
from unittest import TestCase, main

from readalongs.align import create_input_tei
from readalongs.app import app
from readalongs.cli import prepare
from readalongs.log import LOGGER


class TestPrepareCli(TestCase):
    LOGGER.setLevel("DEBUG")
    data_dir = os.path.join(os.path.dirname(__file__), "data")

    def setUp(self):
        app.logger.setLevel("DEBUG")
        self.runner = app.test_cli_runner()
        self.tempdirobj = tempfile.TemporaryDirectory(
            prefix="tmpdir_test_prepare_cli_", dir="."
        )
        self.tempdir = self.tempdirobj.name
        # Alternative tempdir code keeps it after running, for manual inspection:
        # self.tempdir = tempfile.mkdtemp(prefix="tmpdir_test_prepare_cli_", dir=".")
        # print("tmpdir={}".format(self.tempdir))
        self.empty_file = os.path.join(self.tempdir, "empty.txt")
        with io.open(self.empty_file, "wb"):
            pass

    def tearDown(self):
        self.tempdirobj.cleanup()

    def test_invoke_prepare(self):
        results = self.runner.invoke(
            prepare,
            ["-l", "atj", "-d", self.empty_file, os.path.join(self.tempdir, "delme")],
        )
        self.assertEqual(results.exit_code, 0)
        self.assertRegex(results.stdout, "Running readalongs prepare")
        # print('Prepare.stdout: {}'.format(results.stdout))

    def test_no_lang(self):
        results = self.runner.invoke(
            prepare, [self.empty_file, self.empty_file + ".xml"]
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertRegex(results.stdout, "Missing.*language")

    def test_inputfile_not_exist(self):
        results = self.runner.invoke(prepare, "-l atj /file/does/not/exist delme")
        self.assertNotEqual(results.exit_code, 0)
        self.assertRegex(results.stdout, "No such file or directory")

    def test_outputfile_exists(self):
        results = self.runner.invoke(
            prepare,
            ["-l", "atj", self.empty_file, os.path.join(self.tempdir, "exists")],
        )
        results = self.runner.invoke(
            prepare,
            ["-l", "atj", self.empty_file, os.path.join(self.tempdir, "exists")],
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertRegex(results.stdout, "exists.*overwrite")

    def test_output_exists(self):
        xmlfile = os.path.join(self.tempdir, "fra.xml")
        results = self.runner.invoke(
            prepare, ["-l", "fra", os.path.join(self.data_dir, "fra.txt"), xmlfile]
        )
        self.assertEqual(results.exit_code, 0)
        self.assertTrue(os.path.exists(xmlfile), "output xmlfile did not get created")

    def test_output_correct(self):
        input_file = os.path.join(self.data_dir, "fra.txt")
        xml_file = os.path.join(self.tempdir, "fra.xml")
        results = self.runner.invoke(prepare, ["-l", "fra", input_file, xml_file])
        self.assertEqual(results.exit_code, 0)

        ref_file = os.path.join(self.data_dir, "fra-prepared.xml")
        with open(xml_file) as output_f, open(ref_file) as ref_f:
            self.maxDiff = None
            self.assertListEqual(
                list(output_f),
                list(ref_f),
                f"output {xml_file} and reference {ref_file} differ.",
            )

    def test_input_is_stdin(self):
        results = self.runner.invoke(prepare, "-l fra -", input="Ceci est un test.")
        # LOGGER.warning("Output: {}".format(results.output))
        # LOGGER.warning("Exception: {}".format(results.exception))
        self.assertEqual(results.exit_code, 0)
        self.assertIn("<s>Ceci est un test", results.stdout)
        self.assertIn('<text xml:lang="fra">', results.stdout)

    def test_generate_output_name(self):
        input_file = os.path.join(self.tempdir, "someinput.txt")
        copyfile(os.path.join(self.data_dir, "fra.txt"), input_file)
        results = self.runner.invoke(prepare, ["-l", "fra", input_file])
        LOGGER.warning("Output: {}".format(results.output))
        LOGGER.warning("Exception: {}".format(results.exception))
        self.assertEqual(results.exit_code, 0)
        self.assertRegex(results.stdout, "Wrote.*someinput[.]xml")
        self.assertTrue(os.path.exists(os.path.join(self.tempdir, "someinput.xml")))

    def test_prepare_with_different_newlines(self):
        sent = "Ceci est une phrase."
        linux_file = os.path.join(self.tempdir, "linux_file")
        with open(linux_file, mode="wb") as f:
            file_contents = sent + "\n\n" + sent + "\n\n\n" + sent + "\n"
            f.write(file_contents.encode("ascii"))
        linux_results = self.runner.invoke(prepare, ["-l", "fra", linux_file, "-"])
        linux_output = linux_results.output
        # The "Linux" output is the reference output, but we validate it a bit
        # too, with a regex: it has to have 2 pages and 3 paragraphs
        self.assertRegex(
            linux_output,
            re.compile(
                '<div type="page">.*<p>.*<p>.*<div type="page">.*<p>', re.DOTALL
            ),
        )

        no_eol_file = os.path.join(self.tempdir, "no_eol")
        with open(no_eol_file, mode="wb") as f:
            file_contents = sent + "\n\n" + sent + "\n\n\n" + sent
            f.write(file_contents.encode("ascii"))
        no_eol_results = self.runner.invoke(prepare, ["-l", "fra", no_eol_file, "-"])
        no_eol_output = no_eol_results.output
        self.assertEqual(
            linux_output,
            no_eol_output,
            "An absent final newline should not affect prepare",
        )

        dos_file = os.path.join(self.tempdir, "dos_file")
        with open(dos_file, mode="wb") as f:
            file_contents = sent + "\r\n\r\n" + sent + "\r\n\r\n\r\n" + sent + "\r\n"
            f.write(file_contents.encode("ascii"))
        dos_results = self.runner.invoke(prepare, ["-l", "fra", dos_file, "-"])
        dos_output = dos_results.output
        self.assertEqual(
            linux_output,
            dos_output,
            "Using DOS-style newlines should not affect prepare",
        )

        mac_file = os.path.join(self.tempdir, "mac_file")
        with open(mac_file, mode="wb") as f:
            file_contents = sent + "\r\r" + sent + "\r\r\r" + sent + "\r"
            f.write(file_contents.encode("ascii"))
        mac_results = self.runner.invoke(prepare, ["-l", "fra", mac_file, "-"])
        mac_output = mac_results.output
        self.assertEqual(
            linux_output,
            mac_output,
            "Using old Mac-style newlines should not affect prepare",
        )

    def test_create_input_tei_no_input(self):
        with self.assertRaises(RuntimeError):
            (fh, fname) = create_input_tei()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###################################################################
#
# make_smil.py
#
#   Turns alignment into formatted SMIL for ReadAlongs WebComponent
####################################################################


from __future__ import absolute_import, division, print_function, unicode_literals

import argparse

import pystache

from readalongs.text.util import save_txt

SMIL_TEMPLATE = """<smil xmlns="http://www.w3.org/ns/SMIL" version="3.0">
    <body>
        {{#words}}
        <par id="par-{{id}}">
            <text src="{{text_path}}#{{id}}"/>
            <audio src="{{audio_path}}" clipBegin="{{start}}" clipEnd="{{end}}"/>
        </par>
        {{/words}}
    </body>
</smil>
"""

BASENAME_IDX = 0
START_TIME_IDX = 9
WORDS_IDX = 10
WORD_SPAN = 4
WORD_SUBIDX = 2
END_SUBIDX = 3


def parse_hypseg(text):
    """ Parse hypseg alignments file and return alignements

    Parameters
    ----------
    text : str
        hypseg text

    Returns
    -------
    dict
        returns a dictionary of all start and end points for each word in text
    """
    results = {"words": []}
    tokens = text.strip().split()
    # results["basename"] = tokens[BASENAME_IDX]
    start = float(tokens[START_TIME_IDX]) * 0.01
    i = WORDS_IDX
    while i < len(tokens):
        word = tokens[i + WORD_SUBIDX]
        end = tokens[i + END_SUBIDX]
        end = float(end) * 0.01
        if word != "<sil>":
            results["words"].append({"id": word, "start": start, "end": end})
        start = end
        i += WORD_SPAN
    return results


def make_smil(text_path: str, audio_path: str, results: dict) -> str:
    """ Actually render the SMIL

    Parameters
    ----------
    text_path : str
        path to text
    audio_path : str
        path to audio
    results : dict
        all alignements

    Returns
    -------
    str
        formatted SMIL
    """
    results["text_path"] = text_path
    results["audio_path"] = audio_path
    return pystache.render(SMIL_TEMPLATE, results)


def go(seg_path, text_path, audio_path, output_path):
    results = make_smil(text_path, audio_path, parse_hypseg(seg_path))
    save_txt(output_path, results)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert XML to another orthography while preserving tags"
    )
    parser.add_argument("input_seg", type=str, help="Input hypseg file")
    parser.add_argument("text_path", type=str, help="Text filename")
    parser.add_argument("audio_path", type=str, help="Audio filename")
    parser.add_argument("output", type=str, help="Output SMIL file")
    args = parser.parse_args()
    go(args.input_seg, args.text_path, args.audio_path, args.output)



#!/usr/bin/env python3

import os
import tempfile
from unittest import TestCase, main

from lxml import etree

from readalongs.app import app
from readalongs.log import LOGGER
from readalongs.text import tokenize_xml
from readalongs.text.add_ids_to_xml import add_ids


class TestDNAText(TestCase):
    LOGGER.setLevel("DEBUG")
    data_dir = os.path.join(os.path.dirname(__file__), "data")
    maxDiff = None

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_tok_all_words(self):
        txt = """<document xml:lang="fra">
<s>Bonjour! Comment ça va?</s>
<s>Voici une deuxième phrase.</s>
</document>"""
        xml = etree.fromstring(txt)
        tokenized = tokenize_xml.tokenize_xml(xml)
        as_txt = etree.tounicode(tokenized)
        # print(etree.tounicode(tokenized))

        ref = """<document xml:lang="fra">
<s><w>Bonjour</w>! <w>Comment</w> <w>ça</w> <w>va</w>?</s>
<s><w>Voici</w> <w>une</w> <w>deuxième</w> <w>phrase</w>.</s>
</document>"""
        # print('as_txt="' + as_txt +'"')
        # print('ref="' + ref +'"')
        self.assertEqual(as_txt, ref)

        with_ids = add_ids(tokenized)
        ids_as_txt = etree.tounicode(with_ids)
        # print('with ids="' + ids_as_txt + '"')
        ref_with_ids = """<document xml:lang="fra">
<s id="s0"><w id="s0w0">Bonjour</w>! <w id="s0w1">Comment</w> <w id="s0w2">ça</w> <w id="s0w3">va</w>?</s>
<s id="s1"><w id="s1w0">Voici</w> <w id="s1w1">une</w> <w id="s1w2">deuxième</w> <w id="s1w3">phrase</w>.</s>
</document>"""
        self.assertEqual(ids_as_txt, ref_with_ids)

    def test_tok_some_words(self):
        txt = """<document xml:lang="fra">
<p><s>Bonjour! Comment ça va?</s></p>
<p do-not-align="true"><s>Bonjour! Comment ça va?</s></p>
<s do-not-align="TRUE">Voici une deuxième phrase.</s>
<s>Un <foo do-not-align="1">mot ou deux</foo> à exclure.</s>
</document>"""
        xml = etree.fromstring(txt)
        tokenized = tokenize_xml.tokenize_xml(xml)
        as_txt = etree.tounicode(tokenized)
        # print('as_txt="' + as_txt +'"')

        ref = """<document xml:lang="fra">
<p><s><w>Bonjour</w>! <w>Comment</w> <w>ça</w> <w>va</w>?</s></p>
<p do-not-align="true"><s>Bonjour! Comment ça va?</s></p>
<s do-not-align="TRUE">Voici une deuxième phrase.</s>
<s><w>Un</w> <foo do-not-align="1">mot ou deux</foo> <w>à</w> <w>exclure</w>.</s>
</document>"""
        self.assertEqual(as_txt, ref)

        with_ids = add_ids(tokenized)
        ids_as_txt = etree.tounicode(with_ids)
        # print('with ids="' + ids_as_txt + '"')
        ref_with_ids = """<document xml:lang="fra">
<p id="p0"><s id="p0s0"><w id="p0s0w0">Bonjour</w>! <w id="p0s0w1">Comment</w> <w id="p0s0w2">ça</w> <w id="p0s0w3">va</w>?</s></p>
<p do-not-align="true"><s>Bonjour! Comment ça va?</s></p>
<s do-not-align="TRUE">Voici une deuxième phrase.</s>
<s id="s0"><w id="s0w0">Un</w> <foo do-not-align="1">mot ou deux</foo> <w id="s0w1">à</w> <w id="s0w2">exclure</w>.</s>
</document>"""
        self.assertEqual(ids_as_txt, ref_with_ids)

    def test_tok_div_p_s(self):
        txt = """<document xml:lang="fra">
<div>
<p> <s>Une phrase.</s> </p>
<p> <s>Deux phrases.</s> </p>
</div>
<div do-not-align="TRUE">
<p> <s>Une phrase.</s> </p>
<p> <s>Deux phrases.</s> </p>
</div>
<div>
<p do-not-align="1"> <s>Une phrase.</s> </p>
<p> <s do-not-align="true">Deux phrases.</s> </p>
<p> <s>Trois phrases.</s> </p>
</div>
</document>"""
        xml = etree.fromstring(txt)
        tokenized = tokenize_xml.tokenize_xml(xml)
        as_txt = etree.tounicode(tokenized)
        # print('as_txt="' + as_txt +'"')

        ref = """<document xml:lang="fra">
<div>
<p> <s><w>Une</w> <w>phrase</w>.</s> </p>
<p> <s><w>Deux</w> <w>phrases</w>.</s> </p>
</div>
<div do-not-align="TRUE">
<p> <s>Une phrase.</s> </p>
<p> <s>Deux phrases.</s> </p>
</div>
<div>
<p do-not-align="1"> <s>Une phrase.</s> </p>
<p> <s do-not-align="true">Deux phrases.</s> </p>
<p> <s><w>Trois</w> <w>phrases</w>.</s> </p>
</div>
</document>"""
        self.assertEqual(as_txt, ref)

        with_ids = add_ids(tokenized)
        ids_as_txt = etree.tounicode(with_ids)
        # print('with ids="' + ids_as_txt + '"')

        ref_with_ids = """<document xml:lang="fra">
<div id="d0">
<p id="d0p0"> <s id="d0p0s0"><w id="d0p0s0w0">Une</w> <w id="d0p0s0w1">phrase</w>.</s> </p>
<p id="d0p1"> <s id="d0p1s0"><w id="d0p1s0w0">Deux</w> <w id="d0p1s0w1">phrases</w>.</s> </p>
</div>
<div do-not-align="TRUE">
<p> <s>Une phrase.</s> </p>
<p> <s>Deux phrases.</s> </p>
</div>
<div id="d1">
<p do-not-align="1"> <s>Une phrase.</s> </p>
<p id="d1p0"> <s do-not-align="true">Deux phrases.</s> </p>
<p id="d1p1"> <s id="d1p1s0"><w id="d1p1s0w0">Trois</w> <w id="d1p1s0w1">phrases</w>.</s> </p>
</div>
</document>"""
        self.assertEqual(ids_as_txt, ref_with_ids)

    def test_dna_word(self):
        txt = """<s xml:lang="fra">Une <w do-not-align="true">exclude</w> phrase.</s>"""
        xml = etree.fromstring(txt)
        tokenized = tokenize_xml.tokenize_xml(xml)
        self.assertRaises(RuntimeError, add_ids, tokenized)

    def test_dna_word_nested(self):
        txt = """<s xml:lang="fra">Une <foo do-not-align="true"><bar><w>exclude</w></bar></foo> phrase.</s>"""
        xml = etree.fromstring(txt)
        tokenized = tokenize_xml.tokenize_xml(xml)
        self.assertRaises(RuntimeError, add_ids, tokenized)


if __name__ == "__main__":
    main()



"""
Specifies the minimum version of Python required for this project.
"""

import sys

# Please write this as a tuple of (major, minor, patch) verison:
MINIMUM_PYTHON_VERSION_REQUIRED = (3, 7, 0)


def ensure_using_supported_python_version():
    if sys.version_info < MINIMUM_PYTHON_VERSION_REQUIRED:
        sys.exit(
            "Python "
            + _version_to_str(MINIMUM_PYTHON_VERSION_REQUIRED)
            + " required (you are currently using Python "
            + _version_to_str(sys.version_info)
            + "). Please use a newer version of Python."
        )


def _version_to_str(version):
    major, minor, patch = version[:3]
    return "{major}.{minor}.{patch}".format(**locals())


ensure_using_supported_python_version()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
#
# create_epub.py
#
######################################################################

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil

import pystache

from readalongs.log import LOGGER
from readalongs.text.util import (
    copy_file_to_zip,
    ensure_dirs,
    load_txt,
    load_xml,
    load_xml_with_encoding,
    save_txt,
    save_txt_zip,
    xpath_default,
)

EPUB_PATH = "EPUB"
RESOURCES = os.path.dirname(__file__)
MIMETYPE_ORIGIN_PATH = os.path.join(RESOURCES, "templates/mimetype")
MIMETYPE_DEST_PATH = "mimetype"
CONTAINER_ORIGIN_PATH = os.path.join(RESOURCES, "templates/container.xml")
CONTAINER_DEST_PATH = "META-INF/container.xml"
PACKAGE_ORIGIN_PATH = os.path.join(RESOURCES, "templates/package.opf")
PACKAGE_DEST_PATH = os.path.join(EPUB_PATH, "package.opf")
STYLESHEET_ORIGIN_PATH = os.path.join(RESOURCES, "templates/stylesheet.css")
STYLESHEET_DEST_PATH = os.path.join(EPUB_PATH, "stylesheet.css")


def process_src_attrib(src_text, id_prefix, mimetypes):
    filename = src_text.split("#")[0]
    filename_without_ext, ext = os.path.splitext(filename)
    ext = ext.strip(".")
    if ext not in mimetypes:
        LOGGER.warning("Unknown extension in SMIL: %s", ext)
        return None
    entry = {
        "origin_path": filename,
        "dest_path": filename,
        "ext": ext.lower(),
        "id": id_prefix + os.path.basename(filename_without_ext),
        "mimetype": mimetypes[ext],
    }
    return entry


def extract_files_from_SMIL(input_path):
    smil = load_xml(input_path)
    found_files = {}
    xhtml_ids = []
    dirname = os.path.dirname(input_path)

    # add media referenced in the SMIL file itself
    queries = [
        {
            "xpath": ".//i:text/@src",
            "id_prefix": "",
            "mimetypes": {"xhtml": "application/xhtml+xml"},
        },
        {
            "xpath": ".//i:audio/@src",
            "id_prefix": "audio-",
            "mimetypes": {"wav": "audio/wav", "mp3": "audio/mpeg"},
        },
    ]

    for query in queries:
        for src_text in xpath_default(smil, query["xpath"]):
            entry = process_src_attrib(src_text, query["id_prefix"], query["mimetypes"])
            if entry is not None and entry["origin_path"] not in found_files:
                if entry["mimetype"] == "application/xhtml+xml":
                    entry["overlay"] = 'media-overlay="overlay"'
                    xhtml_ids.append({"id": entry["id"]})
                found_files[entry["origin_path"]] = entry

    # add media referenced within the xhtml files (e.g. imgs)
    within_xhtml_queries = [
        {
            "xpath": ".//i:img/@src",
            "id_prefix": "img-",
            "mimetypes": {
                "png": "image/png",
                "jpg": "image/jpeg",
                "jpeg": "image/jpeg",
                "gif": "image/gif",
            },
        }
    ]

    SEARCHABLE_EXTENSIONS = ["xhtml"]
    for entry in found_files.values():
        if entry["ext"] not in SEARCHABLE_EXTENSIONS:
            continue
        origin_path = os.path.join(dirname, entry["origin_path"])
        xhtml = load_xml_with_encoding(origin_path)
        for query in within_xhtml_queries:
            for src_text in xpath_default(xhtml, query["xpath"]):
                entry = process_src_attrib(
                    src_text, query["id_prefix"], query["mimetypes"]
                )
                if entry is not None and entry["origin_path"] not in found_files:
                    found_files[entry["origin_path"]] = entry

    # add this file
    found_files[input_path] = {
        "origin_path": input_path,
        "dest_path": os.path.basename(input_path),
        "id": "overlay",
        "mimetype": "application/smil+xml",
        "ext": "smil",
    }

    return {"media": found_files.values(), "xhtml": xhtml_ids}


def copy_file_to_dir(output_path, origin_path, dest_path):
    """Copy file to a directory, mimicking the interface of
    copy_file_to_zip."""
    shutil.copy(origin_path, os.path.join(output_path, dest_path))


def save_txt_to_dir(output_path, dest_path, txt):
    """Save text to a directory, mimicking the interface of
    save_txt_zip."""
    save_txt(os.path.join(output_path, dest_path), txt)


def create_epub(input_path, output_path, unpacked=False):
    if os.path.isdir(output_path):
        shutil.rmtree(output_path)
    ensure_dirs(output_path)
    input_dirname = os.path.dirname(input_path)
    if unpacked:
        os.mkdir(output_path)
        copy = copy_file_to_dir
        save = save_txt_to_dir
    else:
        copy = copy_file_to_zip
        save = save_txt_zip

    # mimetype file
    copy(output_path, MIMETYPE_ORIGIN_PATH, MIMETYPE_DEST_PATH)

    # container.xml file
    container_template = load_txt(CONTAINER_ORIGIN_PATH)
    container_txt = pystache.render(
        container_template, {"package_path": PACKAGE_DEST_PATH}
    )
    save(output_path, CONTAINER_DEST_PATH, container_txt)

    # the SMIL and all the files referenced in the SMIL
    package_data = extract_files_from_SMIL(input_path)
    package_template = load_txt(PACKAGE_ORIGIN_PATH)
    package_txt = pystache.render(package_template, package_data)
    save(output_path, PACKAGE_DEST_PATH, package_txt)

    for entry in package_data["media"]:
        origin_path = os.path.join(input_dirname, entry["origin_path"])
        if not os.path.exists(origin_path):
            LOGGER.warning("Cannot find file %s to copy into EPUB file", origin_path)
            continue
        dest_path = os.path.join(EPUB_PATH, entry["dest_path"])
        copy(output_path, origin_path, dest_path)

    # CSS file
    copy(output_path, STYLESHEET_ORIGIN_PATH, STYLESHEET_DEST_PATH)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###################################################
#
# pitch2svg.py
#
# Make an SVG of a pitch trace.
#
# Based loosely off code from Martijn Millecamp
# (martijn.millecamp@student.kuleuven.be) and
# Miroslav Masat (miro.masat@gmail.com):
#  https://github.com/miromasat/pitch-detection-librosa-python
#
###################################################

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import os
from io import open
from math import floor

import librosa
import numpy as np
import pystache

from readalongs.log import LOGGER
from readalongs.waveform2svg.audio_util import (
    SAMPLE_RATE,
    load_wav_or_smil,
    save_txt,
    smooth,
)

FMIN = 80
FMAX = 1000
THRESHOLD = 0.75

SVG_TEMPLATE = """<svg id='pitch' preserveAspectRatio='none' viewBox="0 0 {{width}} {{height}}" xmlns="http://www.w3.org/2000/svg" height="{{height}}" width="{{width}}">
    <polygon points="{{#points}}{{x}},{{y}} {{/points}}"></polygon>
</svg>
"""


def render_svg(pitches, width=512, height=100, zero_height=5):
    data = {"height": height, "width": width, "points": []}
    data["points"].append({"x": 0.0, "y": float(height)})
    data["points"].append({"x": 0.0, "y": float(height - zero_height)})
    for i, pitch in enumerate(pitches):
        x = i + 0.5
        y = (1.0 - pitch) * (height - zero_height)
        y = "%.2f" % y
        data["points"].append({"x": x, "y": y})
    data["points"].append({"x": float(width), "y": float(height - zero_height)})
    data["points"].append({"x": float(width), "y": float(height)})
    return pystache.render(SVG_TEMPLATE, data)


def extract_pitches(waveform, nbuckets=512):
    nsamples = waveform.shape[0]
    hop_length = int(floor(nsamples / nbuckets))
    pitches, magnitudes = librosa.core.piptrack(
        y=waveform,
        sr=SAMPLE_RATE,
        fmin=FMIN,
        fmax=FMAX,
        hop_length=hop_length,
        threshold=THRESHOLD,
    )
    pitches = pitches[
        :, :nbuckets
    ]  # TODO: AP: I'm not sure what this is meant to be. Causing error. Maybe just :nbuckets?
    pitches = pitches.max(axis=0)
    pitches /= pitches.max()
    return smooth(pitches, window_size=int(floor(nbuckets / 40)))


def make_pitch_svg(input_path, nbuckets=512, height=100, width=512, zero_height=5):
    waveform = load_wav_or_smil(input_path)
    pitches = extract_pitches(waveform, nbuckets)
    return render_svg(pitches, width, height, zero_height)


def main(input_path, output_path, nbuckets=512, width=512, height=100, zero_height=5):
    svg = make_pitch_svg(input_path, nbuckets, width, height, zero_height)
    save_txt(output_path, svg)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert a WAV or SMIL file to a SVG file of its pitch trace"
    )
    parser.add_argument("input", type=str, help="Input WAV or SMIL file")
    parser.add_argument("output", type=str, help="Output SVG file")
    parser.add_argument(
        "--nbuckets",
        type=int,
        default=512,
        help="Number of sample buckets (default: 256)",
    )
    parser.add_argument(
        "--width", type=int, default=512, help="Width of output SVG (default: 512)"
    )
    parser.add_argument(
        "--height", type=int, default=100, help="Height of output SVG (default: 100)"
    )
    parser.add_argument(
        "--zero_height", type=int, default=5, help="Padding around zero (default: 5)"
    )
    args = parser.parse_args()
    main(
        args.input,
        args.output,
        args.nbuckets,
        args.width,
        args.height,
        args.zero_height,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# app.py
#
#   Initializes a Flask web application with
#   session manager and SocketIO integration.
#
#######################################################################

import os

from flask import Flask
from flask_socketio import SocketIO

from flask_session import Session
from readalongs.python_version import ensure_using_supported_python_version

ensure_using_supported_python_version()

app = Flask(__name__)
app.secret_key = os.urandom(24)
app.config["MAX_CONTENT_LENGTH"] = 16 * 1024 * 1024
app.config["SESSION_TYPE"] = "filesystem"
Session(app)
socketio = SocketIO(app, manage_session=False)

import readalongs.views  # noqa: E402



#!/usr/bin/env python3

import io
import os
import sys
import tempfile
from shutil import copyfile
from unittest import TestCase, main

from readalongs.app import app
from readalongs.cli import prepare, tokenize
from readalongs.log import LOGGER


class TestTokenizeCli(TestCase):
    LOGGER.setLevel("DEBUG")
    data_dir = os.path.join(os.path.dirname(__file__), "data")

    # Set this to True to keep the temp dirs after running, for manual inspection
    # but please don't push a commit setting this to True!
    keep_temp_dir_after_running = False

    def setUp(self):
        app.logger.setLevel("DEBUG")
        self.runner = app.test_cli_runner()
        if not self.keep_temp_dir_after_running:
            self.tempdirobj = tempfile.TemporaryDirectory(
                prefix="tmpdir_test_tokenize_cli_", dir="."
            )
            self.tempdir = self.tempdirobj.name
        else:
            # Alternative tempdir code keeps it after running, for manual inspection:
            self.tempdir = tempfile.mkdtemp(prefix="tmpdir_test_tokenize_cli_", dir=".")
            print("tmpdir={}".format(self.tempdir))

        self.xmlfile = os.path.join(self.tempdir, "fra.xml")
        _ = self.runner.invoke(
            prepare, ["-l", "fra", os.path.join(self.data_dir, "fra.txt"), self.xmlfile]
        )

    def tearDown(self):
        if not self.keep_temp_dir_after_running:
            self.tempdirobj.cleanup()

    def test_invoke_tok(self):
        results = self.runner.invoke(
            tokenize, [self.xmlfile, os.path.join(self.tempdir, "delme")]
        )
        self.assertEqual(results.exit_code, 0)
        self.assertTrue(os.path.exists(os.path.join(self.tempdir, "delme.xml")))

    def test_generate_output_name(self):
        results = self.runner.invoke(tokenize, ["--debug", self.xmlfile])
        self.assertEqual(results.exit_code, 0)
        self.assertTrue(os.path.exists(os.path.join(self.tempdir, "fra.tokenized.xml")))

    def test_with_stdin(self):
        with io.open(self.xmlfile) as f:
            inputtext = f.read()
        results = self.runner.invoke(tokenize, "-", input=inputtext)
        self.assertEqual(results.exit_code, 0)
        self.assertIn(
            "<s><w>Ceci</w> <w>est</w> <w>une</w> <w>phrase</w>", results.output
        )

    def test_file_already_exists(self):
        results = self.runner.invoke(tokenize, [self.xmlfile, self.xmlfile])
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("use -f to overwrite", results.output)

    def test_bad_input(self):
        results = self.runner.invoke(tokenize, "- -", input="this is not XML!")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Error parsing", results.output)
        # LOGGER.warning("Output: {}".format(results.output))
        # LOGGER.warning("Exception: {}".format(results.exception))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# views.py
#
#   Views for ReadAlong Studio web application
#   Interactions are described as websocket events and responses
#   Corresponding JavaScript is found in readalongs/static/js/main.js
#
#######################################################################

import io
import os
from datetime import datetime
from pathlib import Path
from subprocess import run
from tempfile import mkdtemp
from zipfile import ZipFile

import g2p.mappings.langs as g2p_langs
from flask import abort, redirect, render_template, request, send_file, session, url_for
from flask_socketio import emit
from networkx import has_path

from readalongs.app import app, socketio
from readalongs.log import LOGGER

# LANGS_AVAILABLE in g2p lists langs inferred by the directory structure of
# g2p/mappings/langs, but in ReadAlongs, we need all input languages to any mappings.
# E.g., for Michif, we need to allow crg-dv and crg-tmd, but not crg, which is what
# LANGS_AVAILABLE contains. So we define our own list of languages here.
LANGS_AVAILABLE = []

# Set up LANG_NAMES hash table for studio UI to
# properly name the dropdown options
LANG_NAMES = {"eng": "English"}

for k, v in g2p_langs.LANGS.items():
    for mapping in v["mappings"]:
        # add mapping to names hash table
        LANG_NAMES[mapping["in_lang"]] = mapping["language_name"]
        # add input id to all available langs list
        if mapping["in_lang"] not in LANGS_AVAILABLE:
            LANGS_AVAILABLE.append(mapping["in_lang"])

# get the key from all networks in g2p module that have a path to 'eng-arpabet',
# which is needed for the readalongs
# Filter out <lang>-ipa: we only want "normal" input languages.
# Filter out *-norm and crk-no-symbols, these are just intermediate representations.
LANGS = [
    x
    for x in LANGS_AVAILABLE
    if not x.endswith("-ipa")
    and not x.endswith("-equiv")
    and not x.endswith("-no-symbols")
    and g2p_langs.LANGS_NETWORK.has_node(x)
    and has_path(g2p_langs.LANGS_NETWORK, x, "eng-arpabet")
]

# Hack to allow old English LexiconG2P
LANGS += ["eng"]
# Sort LANGS so the -h messages list them alphabetically
LANGS = sorted(LANGS)

ALLOWED_TEXT = ["txt", "xml", "docx"]
ALLOWED_AUDIO = ["wav", "mp3"]
ALLOWED_G2P = ["csv", "xlsx"]
ALLOWED_EXTENSIONS = set(ALLOWED_AUDIO + ALLOWED_G2P + ALLOWED_TEXT)


def allowed_file(filename: str) -> bool:
    """Determines whether filename is allowable

    Parameters
    ----------
    filename : str
        a filename

    Returns
    -------
    bool
        True if allowed
    """
    return "." in filename and filename.rsplit(".", 1)[1].lower() in ALLOWED_EXTENSIONS


def uploaded_files(dir_path: str) -> dict:
    """Returns all files that have been uploaded

    Parameters
    ----------
    dir_path : str
        path to directory where uploaded files are

    Returns
    -------
    dict
        A dictionary containing three keys:
            - audio : A list containing all paths to audio files
            - text  : A list containing all paths to text files
            - maps  : A list containing all paths to mapping files
    """
    upload_dir = Path(dir_path)
    audio = list(upload_dir.glob("*.wav")) + list(upload_dir.glob("*.mp3"))
    text = (
        list(upload_dir.glob("*.txt"))
        + list(upload_dir.glob("*.xml"))
        + list(upload_dir.glob("*.docx"))
    )
    maps = list(upload_dir.glob("*.csv")) + list(upload_dir.glob("*.xlsx"))
    return {
        "audio": [{"path": str(x), "fn": os.path.basename(str(x))} for x in audio],
        "text": [{"path": str(x), "fn": os.path.basename(str(x))} for x in text],
        "maps": [{"path": str(x), "fn": os.path.basename(str(x))} for x in maps],
    }


def update_session_config(**kwargs) -> dict:
    """Update the session configuration for running readalongs aligner.

    Parameters
    ----------
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    dict
        Returns the updated session configuration
    """
    previous_config = session.get("config", {})
    session["config"] = {**previous_config, **kwargs}
    return session["config"]


@app.route("/")
def home():
    """ Home View - go to Step 1 which is for uploading files """
    return redirect(url_for("steps", step=1))


@socketio.on("config update event", namespace="/config")
def update_config(message):
    emit("config update response", {"data": update_session_config(**message)})


@socketio.on("upload event", namespace="/file")
def upload(message):
    if message["type"] == "audio":
        save_path = os.path.join(session["temp_dir"], message["name"])
        session["audio"] = save_path
    if message["type"] == "text":
        save_path = os.path.join(session["temp_dir"], message["name"])
        session["text"] = save_path
    if message["type"] == "mapping":
        save_path = os.path.join(session["temp_dir"], message["name"])
        if "config" in session and "lang" in session["config"]["lang"]:
            del session["config"]["lang"]
        session["mapping"] = save_path
    with open(save_path, "wb") as f:
        f.write(message["data"]["file"])
    emit("upload response", {"data": {"path": save_path}})


# @SOCKETIO.on('remove event', namespace='/file')
# def remove_f(message):
#     path_to_remove = message['data']['path_to_remove']
#     if os.path.exists(path_to_remove) and os.path.isfile(path_to_remove):
#         os.remove(path_to_remove)
#     emit('remove response', {'data': {'removed_file': os.path.basename(path_to_remove)}})

# @SOCKETIO.on('upload event' namespace='file')
# def upload_f(message):


@app.route("/remove", methods=["POST"])
def remove_file():
    if request.method == "POST":
        path = request.data.decode("utf8").split("=")[1]
        os.remove(path)
    return redirect(url_for("steps", step=1))


@app.route("/step/<int:step>")
def steps(step):
    """ Go through steps """
    if step == 1:
        session.clear()
        session["temp_dir"] = mkdtemp()
        temp_dir = session["temp_dir"]
        return render_template(
            "upload.html",
            uploaded=uploaded_files(temp_dir),
            maps=[{"code": m, "name": LANG_NAMES[m]} for m in LANGS],
        )
    elif step == 2:
        return render_template("preview.html")
    elif step == 3:
        if "audio" not in session or "text" not in session:
            log = "Sorry, it looks like something is wrong with your audio or text. Please try again"
        else:
            flags = ["--force-overwrite"]
            for option in ["--closed-captioning", "--save-temps", "--text-grid"]:
                if session["config"].get(option, False):
                    flags.append(option)
            if session["text"].endswith("txt"):
                flags.append("--text-input")
                flags.append("--language")
                flags.append(session["config"]["lang"])
            timestamp = str(int(datetime.now().timestamp()))
            output_base = "aligned" + timestamp
            args = (
                ["readalongs", "align"]
                + flags
                + [
                    session["text"],
                    session["audio"],
                    os.path.join(session["temp_dir"], output_base),
                ]
            )
            LOGGER.warning(args)
            fname, audio_ext = os.path.splitext(session["audio"])
            data = {"audio_ext": audio_ext, "base": output_base}
            if session["config"].get("show-log", False):
                log = run(args, capture_output=True)
                data["log"] = log
            else:
                run(args)
            data["audio_path"] = os.path.join(
                session["temp_dir"], output_base, output_base + audio_ext
            )
            data["audio_fn"] = f"/file/{output_base}" + audio_ext
            data["text_path"] = os.path.join(
                session["temp_dir"], output_base, output_base + ".xml"
            )
            data["text_fn"] = f"/file/{output_base}" + ".xml"
            data["smil_path"] = os.path.join(
                session["temp_dir"], output_base, output_base + ".smil"
            )
            data["smil_fn"] = f"/file/{output_base}" + ".smil"
        return render_template("export.html", data=data)
    else:
        abort(404)


@app.route("/download/<string:base>", methods=["GET"])
def show_zip(base):
    files_to_download = os.listdir(os.path.join(session["temp_dir"], base))
    if (
        "temp_dir" not in session
        or not os.path.exists(session["temp_dir"])
        or not files_to_download
        or not any(x.startswith("aligned") for x in files_to_download)
    ):
        return abort(
            404, "Nothing to download. Please go to Step 1 of the Read Along Studio"
        )

    data = io.BytesIO()
    with ZipFile(data, mode="w") as z:
        for fname in files_to_download:
            path = os.path.join(session["temp_dir"], base, fname)
            if fname.startswith("aligned"):
                z.write(path, fname)
    data.seek(0)

    if not data:
        return abort(400, "Invalid zip file")

    return send_file(
        data,
        mimetype="application/zip",
        as_attachment=True,
        attachment_filename="data_bundle.zip",
    )


@app.route("/file/<string:fname>", methods=["GET"])
def return_temp_file(fname):
    fn, ext = os.path.splitext(fname)
    LOGGER.warning(session["temp_dir"])
    path = os.path.join(session["temp_dir"], fn, fname)
    if os.path.exists(path):
        return send_file(path)
    else:
        abort(404, "Sorry, we couldn't find that file.")



#!/usr/bin/env python3

import io
import os
import tempfile
from unittest import TestCase, main

from lxml import etree

from readalongs.align import align_audio
from readalongs.app import app
from readalongs.cli import align, g2p, prepare, tokenize
from readalongs.text.convert_xml import convert_xml


class TestG2pCli(TestCase):
    data_dir = os.path.join(os.path.dirname(__file__), "data")

    # Set this to True to keep the temp dirs after running, for manual inspection
    # but please don't push a commit setting this to True!
    keep_temp_dir_after_running = False
    # Set this to True to display the output of many commands invoked here, for building
    # and debugging this test suite
    show_invoke_output = False

    def setUp(self):
        app.logger.setLevel("DEBUG")
        self.runner = app.test_cli_runner()
        if not self.keep_temp_dir_after_running:
            # Temporary directories that get automatically cleaned up:
            self.tempdirobj = tempfile.TemporaryDirectory(
                prefix="tmpdir_test_g2p_cli_", dir="."
            )
            self.tempdir = self.tempdirobj.name
        else:
            # Alternative tempdir code keeps it after running, for manual inspection:
            self.tempdir = tempfile.mkdtemp(prefix="tmpdir_test_g2p_cli_", dir=".")
            print("tmpdir={}".format(self.tempdir))

    def tearDown(self):
        if not self.keep_temp_dir_after_running:
            self.tempdirobj.cleanup()

    def test_invoke_g2p(self):
        input_file = os.path.join(self.data_dir, "fra-tokenized.xml")
        g2p_file = os.path.join(self.tempdir, "fra-g2p.xml")
        results = self.runner.invoke(g2p, [input_file, g2p_file])
        # print(f"g2p results.output='{results.output}'")
        self.assertEqual(results.exit_code, 0)
        self.assertTrue(os.path.exists(os.path.join(self.tempdir, "fra-g2p.xml")))

        # Testing that it fails when the file already exists has to be in the same test,
        # otherwise we have a different tempdir and the file won't already exist
        results = self.runner.invoke(g2p, [input_file, g2p_file])
        # print(f"g2p results.output='{results.output}'")
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("use -f to overwrite", results.output)

        # And add -f to force the overwrite
        results = self.runner.invoke(g2p, ["-f", input_file, g2p_file])
        # print(f"g2p results.output='{results.output}'")
        self.assertEqual(results.exit_code, 0)

    def test_bad_fallback_lang(self):
        input_file = os.path.join(self.data_dir, "fra-tokenized.xml")
        results = self.runner.invoke(
            g2p, ["--g2p-fallback=fra:notalang:und", input_file, "-"]
        )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Invalid value: g2p fallback lang", results.output)

    def test_bad_xml_input(self):
        input_file = os.path.join(self.data_dir, "ej-fra.txt")
        results = self.runner.invoke(g2p, ["--debug", input_file, "-"])
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("Error parsing input file", results.output)

    def test_mixed_langs(self):
        input_file = os.path.join(self.data_dir, "mixed-langs.tokenized.xml")
        g2p_file = os.path.join(self.tempdir, "mixed-langs.g2p.xml")
        results = self.runner.invoke(g2p, [input_file, g2p_file])
        # print(f"g2p results.output='{results.output}'")
        self.assertEqual(results.exit_code, 0)
        self.assertTrue(os.path.exists(g2p_file))

        ref_file = os.path.join(self.data_dir, "mixed-langs.g2p.xml")
        with open(g2p_file) as output_f, open(ref_file) as ref_f:
            self.maxDiff = None
            self.assertListEqual(
                list(output_f),
                list(ref_f),
                f"output {g2p_file} and reference {ref_file} differ.",
            )

    # Write text to a temp file, pass it through prepare -l lang, and then tokenize,
    # saving the final results into filename.
    # filename is assumed to be inside self.tempdir, so we count on tearDown() to clean up.
    def write_prepare_tokenize(self, text, lang, filename):
        with open(filename + ".input.txt", "w", encoding="utf8") as f:
            print(text, file=f)
        self.runner.invoke(
            prepare, ["-l", lang, filename + ".input.txt", filename + ".prepared.xml"]
        )
        self.runner.invoke(tokenize, [filename + ".prepared.xml", filename])

    def test_english_oov(self):
        tok_file = os.path.join(self.tempdir, "tok.xml")
        self.write_prepare_tokenize("This is a froobnelicious OOV.", "eng", tok_file)
        results = self.runner.invoke(g2p, [tok_file])
        if self.show_invoke_output:
            print(
                f"test_english_oov: g2p "
                f"results.output='{results.output}' "
                f"results.exception={results.exception!r}"
            )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("could not be g2p", results.output)
        # self.assertTrue(isinstance(results.exception, KeyError))

        # with a fall back to und, it works
        results = self.runner.invoke(g2p, ["--g2p-fallback", "und", tok_file, "-"])
        if self.show_invoke_output:
            print(
                f"test_english_oov with fallback: g2p "
                f"results.output='{results.output}' "
                f"results.exception={results.exception!r}"
            )
        self.assertEqual(results.exit_code, 0)

    def test_french_oov(self):
        tok_file = os.path.join(self.tempdir, "tok.xml")
        g2p_file = os.path.join(self.tempdir, "g2p.xml")
        self.write_prepare_tokenize(
            "Le ñ n'est pas dans l'alphabet français.", "fra", tok_file
        )
        results = self.runner.invoke(g2p, [tok_file, g2p_file])
        if self.show_invoke_output:
            print(
                f"test_french_oov: g2p "
                f"results.output='{results.output}' "
                f"results.exception={results.exception!r}"
            )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("could not be g2p", results.output)

        # with a fall back to und, it works
        g2p_file2 = os.path.join(self.tempdir, "g2p-fallback.xml")
        results = self.runner.invoke(
            g2p, ["--g2p-fallback", "und", tok_file, g2p_file2]
        )
        if self.show_invoke_output:
            print(
                f"test_french_oov with fallback: g2p "
                f"results.output='{results.output}' "
                f"results.exception={results.exception!r}"
            )
        self.assertEqual(results.exit_code, 0)

    def test_three_way_fallback(self):
        tok_file = os.path.join(self.tempdir, "text.tokenized.xml")
        g2p_file = os.path.join(self.tempdir, "text.g2p.xml")
        self.write_prepare_tokenize(
            "In French été works but Nunavut ᓄᓇᕗᑦ does not.", "eng", tok_file
        )
        # Here we also test generating the output filename from the input filename
        results = self.runner.invoke(g2p, ["--g2p-fallback", "fra:iku", tok_file])
        if self.show_invoke_output:
            print(
                f"test_three_way_fallback: g2p "
                f"results.output='{results.output}' "
                f"results.exception={results.exception!r}"
            )
        self.assertEqual(results.exit_code, 0)
        self.assertTrue(os.path.exists(g2p_file))
        self.assertNotIn("not recognized as IPA", results.output)
        self.assertNotIn("not fully valid eng-arpabet", results.output)

        # Run with verbose output and look for the warning messages
        results = self.runner.invoke(
            g2p,
            ["--g2p-fallback=fra:iku", "--g2p-verbose", tok_file, g2p_file + "verbose"],
        )
        if self.show_invoke_output:
            print(
                f"test_three_way_fallback: g2p "
                f"results.output='{results.output}' "
                f"results.exception={results.exception!r}"
            )
        self.assertEqual(results.exit_code, 0)
        self.assertIn("not recognized as IPA", results.output)
        self.assertIn("not fully valid eng-arpabet", results.output)

        # this text also works with "und", now that we use unidecode
        results = self.runner.invoke(g2p, ["--g2p-fallback=und", tok_file, "-"])
        self.assertEqual(results.exit_code, 0)
        self.assertIn("Trying fallback: und", results.output)

    def test_align_with_error(self):
        text_file = os.path.join(self.tempdir, "input.txt")
        with io.open(text_file, "w", encoding="utf8") as f:
            print("In French été works but Nunavut ᓄᓇᕗᑦ does not.", file=f)
        empty_wav = os.path.join(self.tempdir, "empty.wav")
        with io.open(empty_wav, "wb"):
            pass
        output_dir = os.path.join(self.tempdir, "aligned")
        results = self.runner.invoke(
            align, ["-l", "eng", "-i", text_file, empty_wav, output_dir]
        )
        if self.show_invoke_output:
            print(
                f"align with wrong language: "
                f"results.output='{results.output}' "
                f"results.exception={results.exception!r}"
            )
        self.assertNotEqual(results.exit_code, 0)
        self.assertIn("could not be g2p", results.output)

        results = self.runner.invoke(
            align,
            [
                "-l",
                "eng",
                "-i",
                "-f",
                "--g2p-fallback=fra:iku",
                text_file,
                empty_wav,
                output_dir,
            ],
        )
        if self.show_invoke_output:
            print(
                f"align with wrong language, plus fallback: "
                f"results.output='{results.output}' "
                f"results.exception={results.exception!r}"
            )
        self.assertIn("Trying fallback: fra", results.output)
        self.assertIn("Trying fallback: iku", results.output)
        # We get the error about reading the audio file only if g2p succeeded:
        self.assertIn("Error reading audio file", results.output)

    def test_with_stdin(self):
        input_file = os.path.join(self.data_dir, "fra-tokenized.xml")
        with io.open(input_file, encoding="utf8") as f:
            inputtext = f.read()
        results = self.runner.invoke(g2p, "-", input=inputtext)
        self.assertEqual(results.exit_code, 0)
        self.assertIn("S AH S IY", results.output)

    def test_align_with_invalid_preg2p(self):
        txt = """<document><s xml:lang="und">
            <w>word</w>
            <w ARPABET="G OW D">good</w>
            <w ARPABET="NOT ARPABET">error</w>
        </s></document>"""
        input_file = os.path.join(self.tempdir, "pre-g2p.xml")
        with open(input_file, "w") as f:
            print(txt, file=f)

        results = self.runner.invoke(g2p, [input_file, "-"])
        self.assertNotEqual(results.exit_code, 0)
        # print(results.output)
        self.assertIn("could not be g2p", results.output)
        self.assertIn('<w id="s0w0" ARPABET="W OW D D">word</w>', results.output)
        self.assertIn('<w ARPABET="G OW D" id="s0w1">good</w>', results.output)
        self.assertIn('<w ARPABET="NOT ARPABET" id="s0w2">error</w>', results.output)

        audio_file = os.path.join(self.data_dir, "ej-fra.m4a")
        with self.assertRaises(RuntimeError) as e:
            results = align_audio(input_file, audio_file)
        self.assertIn("could not be g2p'd", str(e.exception))

    def test_align_with_preg2p(self):
        text_file = os.path.join(self.data_dir, "mixed-langs.tokenized.xml")
        audio_file = os.path.join(self.data_dir, "ej-fra.m4a")
        _ = align_audio(
            text_file, audio_file, save_temps=os.path.join(self.tempdir, "foo")
        )
        with open(os.path.join(self.tempdir, "foo.dict"), "r") as f:
            dict_file = f.read()
            self.assertIn("D G IY T UW P IY D", dict_file)

    def run_convert_xml(self, str):
        return etree.tounicode(convert_xml(etree.fromstring(str))[0])

    def test_convert_xml(self):
        self.assertEqual(
            self.run_convert_xml("<t><w>word</w><w></w><n>not word</n></t>"),
            '<t><w ARPABET="W OW D D">word</w><w/><n>not word</n></t>',
        )

        self.assertEqual(
            self.run_convert_xml(
                '<s><w xml:lang="eng">Patrick</w><w xml:lang="kwk-umista">xtła̱n</w></s>'
            ),
            '<s><w xml:lang="eng" ARPABET="P AE T R IH K">Patrick</w><w xml:lang="kwk-umista" ARPABET="K Y T S AH N">xtła̱n</w></s>',
        )

        self.assertEqual(
            self.run_convert_xml('<s><w xml:lang="und">Patrickxtła̱n</w></s>'),
            '<s><w xml:lang="und" ARPABET="P AA T D IY CH K K T L AA N">Patrickxtła̱n</w></s>',
        )

    def test_convert_xml_invalid(self):
        xml = etree.fromstring('<s><w ARPABET="V AA L IY D">valid</w></s>')
        c_xml, valid = convert_xml(xml)
        self.assertEqual(
            etree.tounicode(c_xml), '<s><w ARPABET="V AA L IY D">valid</w></s>'
        )
        self.assertTrue(valid, "convert_xml with valid pre-g2p'd text")

        xml = etree.fromstring('<s><w ARPABET="invalid">invalid</w></s>')
        c_xml, valid = convert_xml(xml)
        self.assertEqual(
            etree.tounicode(c_xml), '<s><w ARPABET="invalid">invalid</w></s>'
        )
        self.assertFalse(valid, "convert_xml with invalid pre-g2p'd text")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###################################################
#
# pitch2svg.py
#
# Make an SVG of a pitch trace.
#
# Based loosely off code from Martijn Millecamp
# (martijn.millecamp@student.kuleuven.be) and
# Miroslav Masat (miro.masat@gmail.com):
#  https://github.com/miromasat/pitch-detection-librosa-python
#
###################################################

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import os
from collections import OrderedDict
from io import open

import librosa
import numpy as np
import pystache
from audio_util import save_txt
from lxml import etree

from readalongs.log import LOGGER
from readalongs.text.util import xpath_default

FMIN = 80
FMAX = 1000
THRESHOLD = 0.75

SVG_TEMPLATE = """<svg id='units' preserveAspectRatio='none' viewBox="0 0 {{total_width}} {{total_height}}" xmlns="http://www.w3.org/2000/svg" height="{{total_height}}" width="{{total_width}}">
  {{#rects}}
    <rect x="{{x}}" y="0" rx="{{radius}}" ry="{{radius}}" width="{{width}}" height="{{total_height}}"></rect>
  {{/rects}}
</svg>
"""


def render_svg(data, width=512, height=100, radius=4):
    result = {
        "total_width": width,
        "total_height": height,
        "radius": radius,
        "rects": [],
    }
    total_duration = data["duration"]
    for src, audio in data["audio_files"].items():
        for unit in audio["sub_units"]:
            x = (unit["start"] / total_duration) * width - 0.5
            x = "%.2f" % x
            w = max((unit["duration"] / total_duration) * width - 1.0, 1.0)
            w = "%.2f" % w
            result["rects"].append({"x": x, "width": w})
    return pystache.render(SVG_TEMPLATE, result)


def load_xml(input_path):
    with open(input_path, "r", encoding="utf-8") as fin:
        return etree.fromstring(fin.read())


def parse_smil(input_path):
    """ Figure out the overall start and end of every unit, even if the whole
        sequence plays out over multiple audio files """
    xml = load_xml(input_path)
    data = {"audio_files": OrderedDict()}
    dirname = os.path.dirname(input_path)
    current_time = 0.0
    for audio_node in xpath_default(xml, ".//i:audio"):
        src = audio_node.attrib["src"]
        if src not in data["audio_files"]:
            # get some basic info on the audio
            audio_path = os.path.join(dirname, src)
            y, sr = librosa.load(audio_path)
            duration = y.shape[0] / sr
            data["audio_files"][src] = {
                "src": src,
                "start": current_time,
                "duration": duration,
                "end": current_time + duration,
                "sub_units": [],
            }
            current_time += duration
        current_audio_file = data["audio_files"][src]
        start = (
            float(audio_node.attrib["clipBegin"]) + current_audio_file["start"]
            if "clipBegin" in audio_node.attrib
            else current_audio_file["start"]
        )
        end = (
            float(audio_node.attrib["clipEnd"]) + current_audio_file["start"]
            if "clipEnd" in audio_node.attrib
            else current_audio_file["end"]
        )
        current_audio_file["sub_units"].append(
            {"start": start, "duration": end - start}
        )
    if not data["audio_files"]:
        data["duration"] = 0.0
    else:
        last_audio = next(reversed(data["audio_files"]))
        data["duration"] = data["audio_files"][last_audio]["end"]
    return data


def make_units_svg(input_path, width=512, height=100, radius=4):
    data = parse_smil(input_path)
    return render_svg(data, width, height, radius)


def main(input_path, output_path, width=512, height=100, radius=4):
    svg = make_units_svg(input_path, width, height, radius)
    save_txt(output_path, svg)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert a SMIL file to a SVG file indicating sub-unit durations"
    )
    parser.add_argument("input", type=str, help="Input SMIL file")
    parser.add_argument("output", type=str, help="Output SVG file")
    parser.add_argument(
        "--width", type=int, default=512, help="Width of output SVG (default: 512)"
    )
    parser.add_argument(
        "--height", type=int, default=100, help="Height of output SVG (default: 100)"
    )
    parser.add_argument(
        "--radius",
        type=int,
        default=100,
        help="Radius of rounded rectangles (default: 4)",
    )
    args = parser.parse_args()
    main(args.input, args.output, args.width, args.height, args.radius)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###################################################
#
# add_ids_to_xml.py
#
# In order to tell visualization systems, "highlight this
# thing at this time", the document has to be able to identify
# particular elements.  If the original document does NOT have
# id tags on its elements, this module adds some.
#
# The auto-generated IDs have formats like "s0w2m1" meaning
# "sentence 0, word 2, morpheme 1".  But it's flexible if some elements
# already have ids, or if the markup uses different tags than a TEI document.
#
###################################################

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
from collections import defaultdict
from copy import deepcopy

from lxml import etree

from readalongs.text.util import is_do_not_align, load_xml, save_xml

TAG_TO_ID = {
    "text": "t",
    "body": "b",
    "div": "d",
    "page": "pp",
    "p": "p",
    "u": "u",
    "s": "s",
    "w": "w",
    "m": "m",
}

TAGS_TO_IGNORE = ["head", "teiHeader", "script"]


def add_ids_aux(
    element: etree, ids: defaultdict = defaultdict(lambda: 0), parent_id: str = ""
) -> defaultdict:
    """ Add ids to xml element

    Parameters
    ----------
    element : etree
        element to add ids to
    ids : defaultdict, optional
        , by default defaultdict(lambda: 0)
    parent_id : str, optional
        id of parent element, by default ''

    Returns
    -------
    defaultdict
        ids added to element
    """
    if element.tag is etree.Comment:
        return ids
    tag = etree.QName(element.tag).localname
    if tag in TAGS_TO_IGNORE:
        return ids
    if is_do_not_align(element):
        if tag == "w":
            raise RuntimeError(
                'Found <w> element with do-not-align="true" attribute. '
                "This is not allowed, please verify you XML input."
            )
        if element.xpath(".//w"):
            raise RuntimeError(
                'Found <w> nested inside a do-not-align="true" element. '
                "This is not allowed, please verify you XML input."
            )
        return ids
    if "id" not in element.attrib:
        if tag in TAG_TO_ID:
            id = TAG_TO_ID[tag]
        elif tag == "seg" and "type" in element.attrib:
            if element.attrib["type"] == "syll":
                id = "y"
            elif element.attrib["type"] in [
                "morph",
                "morpheme",
                "base",
                "root",
                "prefix",
                "suffix",
            ]:
                id = "m"
        else:
            id = tag
        if id not in ids:
            ids[id] = 0
        element.attrib["id"] = parent_id + id + str(ids[id])
        ids[id] += 1
    full_id = element.attrib["id"]
    new_ids = deepcopy(ids)
    for child in element:
        new_ids = add_ids_aux(child, new_ids, full_id)
    return ids


def add_ids(xml: etree) -> etree:
    """Add ids to xml

    Parameters
    ----------
    xml : etree
        xml to add ids to

    Returns
    -------
    etree
        xml with ids added
    """
    xml = deepcopy(xml)
    ids: defaultdict = defaultdict(lambda: 0)
    for child in xml:  # don't bother with the root element
        if child.tag is etree.Comment:
            continue
        ids = add_ids_aux(child, ids)
    return xml


def go(input_filename: str, output_filename: str) -> None:
    xml = load_xml(input_filename)
    xml = add_ids(xml)
    save_xml(output_filename, xml)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert XML to another orthography while preserving tags"
    )
    parser.add_argument("input", type=str, help="Input XML")
    parser.add_argument("output", type=str, help="Output XML")
    args = parser.parse_args()
    go(args.input, args.output)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

##################################################
#
# make_fsg.py
#
# This module takes a text file, marked up with
# units (e.g. w for word, m for morpheme) and ids
# and converted to IPA, and outputs a FSG
# file for processing by PocketSphinx.
#
##################################################


from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import datetime
import os

import pystache

from readalongs.text.util import load_xml, save_txt

# ###################3
#
# For making an FSG from a SMIL, we have the following rules:
#
# -- All children of a SEQ element are subject to alignment, in the order they
# occur.
#
# -- Only one TEXT child of a PAR element is subject to alignment, with a
# for the first child.
#
# -- a BODY element is treated exactly as a SEQ
#
#
# TODO: AP: Do we need this? It doesn't appear to be used anywhere.
#       There's also an undefined variable error on line 90.
# TODO: Add numpy standard docstrings to functions
#########################


class GrammarComposite:
    def __init__(self, id):
        self.id = id
        self.children = []

    def append(self, child):
        self.children.append(child)

    def get_id_as_str(self):
        return "<%s>" % self.id


class GrammarChoice(GrammarComposite):
    def to_jsgf(self):
        results = []
        child_ids = " | ".join(c.get_id_as_str() for c in self.children)
        results.append("%s = %s" % (self.get_id_as_str(), child_ids))
        for child in self.children:
            results += child.to_jsgf()
        return results


class GrammarSequence(GrammarComposite):
    def to_jsgf(self):
        results = []
        child_ids = " ".join(c.get_id_as_str() for c in self.children)
        results.append("%s = %s" % (self.get_id_as_str(), child_ids))
        for child in self.children:
            results += child.to_jsgf()
        return results


def make_sequence(seq_node):
    for child in seq_node:
        # TODO: flake8 flags child_id as an unused variable, and indeed, this function
        # basically does nothing. Figure out what it's supposed to do and fix this
        # function! -EJ
        child_id = child.attrib["id"]


def make_jsgf(smil, unit="m"):
    body_node = xpath_default(smil, ".//i:body")[0]
    for child in body_node:
        print(child.tag)


def main(input_filename, output_filename, unit):
    smil = load_xml(input_filename)
    jsgf = make_jsgf(smil, unit)
    # save_txt(output_filename, jsgf)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Make an JSGF grammar from an XML file with IDs"
    )
    parser.add_argument("input", type=str, help="Input SMIL")
    parser.add_argument("output_jsgf", type=str, help="Output .jsgf file")
    parser.add_argument(
        "--unit",
        type=str,
        default="m",
        help="XML tag of the unit of analysis " '(e.g. "w" for word, "m" for morpheme)',
    )
    args = parser.parse_args()
    main(args.input, args.output_fsg, args.unit)



#!/usr/bin/env python3

"""
Test force-alignment with PocketSphinx FSG search from Python API
"""

import os
import tempfile
import unittest

from lxml import etree

from readalongs.align import (
    align_audio,
    convert_to_xhtml,
    create_input_tei,
    create_input_xml,
)
from readalongs.log import LOGGER
from readalongs.portable_tempfile import PortableNamedTemporaryFile
from readalongs.text.util import load_txt, save_xml


class TestForceAlignment(unittest.TestCase):
    LOGGER.setLevel("DEBUG")
    data_dir = os.path.join(os.path.dirname(__file__), "data")

    def testAlign(self):
        xml_path = os.path.join(self.data_dir, "ej-fra.xml")
        wav_path = os.path.join(self.data_dir, "ej-fra.m4a")
        results = align_audio(xml_path, wav_path, unit="w")

        # Verify that the same IDs are in the output
        converted_path = os.path.join(self.data_dir, "ej-fra-converted.xml")
        xml = etree.parse(converted_path).getroot()
        words = results["words"]
        xml_words = xml.xpath(".//w")
        self.assertEqual(len(words), len(xml_words))
        for w, xw in zip(words, xml_words):
            self.assertEqual(xw.attrib["id"], w["id"])

    def testAlignText(self):
        txt_path = os.path.join(self.data_dir, "ej-fra.txt")
        wav_path = os.path.join(self.data_dir, "ej-fra.m4a")
        # tempfh, temp_fn = create_input_xml(txt_path, text_language='git', save_temps="unit")
        tempfh, temp_fn = create_input_tei(
            input_file_name=txt_path, text_language="fra", save_temps=None
        )
        results = align_audio(temp_fn, wav_path, unit="w", save_temps=None)

        # Verify that the same IDs are in the output
        converted_path = os.path.join(self.data_dir, "ej-fra-converted.xml")
        xml = etree.parse(converted_path).getroot()
        words = results["words"]
        xml_words = xml.xpath(".//w")
        self.assertEqual(len(words), len(xml_words))
        for w, xw in zip(words, xml_words):
            self.assertEqual(xw.attrib["id"], w["id"])


class TestXHTML(unittest.TestCase):
    data_dir = os.path.join(os.path.dirname(__file__), "data")

    def testConvert(self):
        xml_path = os.path.join(self.data_dir, "ej-fra-converted.xml")
        xml = etree.parse(xml_path).getroot()
        convert_to_xhtml(xml)
        with PortableNamedTemporaryFile(suffix=".xml") as tf:
            save_xml(tf.name, xml)
            txt = load_txt(tf.name)
            self.maxDiff = None
            self.assertEqual(
                txt, load_txt(os.path.join(self.data_dir, "ej-fra-converted.xhtml")),
            )


if __name__ == "__main__":
    LOGGER.setLevel("DEBUG")
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#####################################################################################
#
# run.py
#
#   It seems Werkzeug doesn't play well with SocketIO, so use SocketIO to run the app
#   Watch this issue: https://github.com/miguelgrinberg/Flask-SocketIO/issues/894
#
######################################################################################

from readalongs.app import app, socketio


def run():
    """ Run app using SocketIO
    """
    socketio.run(app)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3

from unittest import TestCase, main

from g2p import make_g2p
from g2p.mappings import Mapping
from g2p.transducer import Transducer

from readalongs.log import LOGGER


class TestIndices(TestCase):
    def setUp(self):
        pass

    def test_basic_composition(self):
        mapping = Mapping([{"in": "a", "out": "b"}])
        transducer = Transducer(mapping)
        tg = transducer("abba")
        self.assertEqual(tg.output_string, "bbbb")
        self.assertEqual(tg.edges, [(0, 0), (1, 1), (2, 2), (3, 3)])

    def test_tiered_composition(self):
        transducer = make_g2p("dan", "eng-arpabet")
        tg = transducer("hej")
        self.assertEqual(tg.output_string, "HH EH Y ")
        self.assertEqual(
            tg.edges,
            [
                [(0, 0), (1, 1), (2, 2)],
                [(0, 0), (1, 1), (2, 2)],
                [(0, 0), (0, 1), (0, 2), (1, 3), (1, 4), (1, 5), (2, 6), (2, 7)],
            ],
        )
        self.assertEqual(
            tg.pretty_edges(),
            [
                [["h", "h"], ["e", "ɛ"], ["j", "j"]],
                [["h", "h"], ["ɛ", "ɛ"], ["j", "j"]],
                [
                    ["h", "H"],
                    ["h", "H"],
                    ["h", " "],
                    ["ɛ", "E"],
                    ["ɛ", "H"],
                    ["ɛ", " "],
                    ["j", "Y"],
                    ["j", " "],
                ],
            ],
        )

    def test_composition_with_none(self):
        transducer = make_g2p("ctp", "eng-arpabet")
        tg = transducer("Qne\u1D2C")
        self.assertEqual(tg.output_string, "HH N EY ")
        self.assertEqual(
            tg.edges,
            [
                [(0, 0), (1, 1), (2, 2), (3, None)],
                [(0, 0), (1, 1), (2, 2), (2, 3)],
                [(0, 0), (0, 1), (0, 2), (1, 3), (1, 4), (2, 5), (3, 6), (3, 7)],
            ],
        )
        self.assertEqual(
            tg.pretty_edges(),
            [
                [["q", "ʔ"], ["n", "n"], ["e", "e"], ["ᴬ", None]],
                [["ʔ", "ʔ"], ["n", "n"], ["e", "e"], ["e", "ː"]],
                [
                    ["ʔ", "H"],
                    ["ʔ", "H"],
                    ["ʔ", " "],
                    ["n", "N"],
                    ["n", " "],
                    ["e", "E"],
                    ["ː", "Y"],
                    ["ː", " "],
                ],
            ],
        )

    def test_fra(self):
        transducer = make_g2p("fra", "eng-arpabet")
        tg = transducer("mais")
        self.assertEqual(tg.output_string, "M EH ")


if __name__ == "__main__":
    LOGGER.setLevel("DEBUG")
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

########################################################
#
# make_all_svgs.py
#
# Take a SMIL file and render all available SVGs for it
#
########################################################

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
from io import open

from audio_util import save_txt
from pitch2svg import make_pitch_svg
from units2svg import make_units_svg
from waveform2svg import make_waveform_svg


def main(
    input_path,
    output_waveform_path,
    output_half_waveform_path,
    output_pitch_path,
    output_units_path,
):
    waveform_svg = make_waveform_svg(input_path)
    save_txt(output_waveform_path, waveform_svg)
    half_waveform_svg = make_waveform_svg(input_path, include_neg=False)
    save_txt(output_half_waveform_path, half_waveform_svg)
    pitch_svg = make_pitch_svg(input_path)
    save_txt(output_pitch_path, pitch_svg)
    units_svg = make_units_svg(input_path)
    save_txt(output_units_path, units_svg)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert a SMIL file to a SVG file of its waveform, pitch, and units"
    )
    parser.add_argument("input", type=str, help="Input SMIL file")
    parser.add_argument("output_waveform", type=str, help="Output SVG file")
    parser.add_argument("output_half_waveform", type=str, help="Output SVG file")
    parser.add_argument("output_pitch", type=str, help="Output SVG file")
    parser.add_argument("output_units", type=str, help="Output SVG file")
    args = parser.parse_args()
    main(
        args.input,
        args.output_waveform,
        args.output_half_waveform,
        args.output_pitch,
        args.output_units,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

##################################################
#
# make_dict.py
#
# This module takes a text file, marked up with
# units (e.g. w for word, m for morpheme) and ids
# and converted to IPA, and produces a
# .dict file for processing by PocketSphinx.
#
##################################################


from __future__ import absolute_import, division, print_function, unicode_literals

import argparse

import pystache

from readalongs.log import LOGGER
from readalongs.text.util import load_xml, save_txt

DICT_TEMPLATE = """{{#items}}
{{id}}\t{{pronunciation}}
{{/items}}
"""


def make_dict(xml, input_filename, unit="m"):
    data = {"items": []}
    nwords = 0
    for e in xml.xpath(".//" + unit):
        if "id" not in e.attrib:
            LOGGER.error(
                "%s-type element without id in file %s" % (unit, input_filename)
            )
        text = e.attrib.get("ARPABET", "").strip()
        if not text:
            continue
        nwords += 1
        data["items"].append({"id": e.attrib["id"], "pronunciation": text})
    if nwords == 0:
        raise RuntimeError("No words in dictionary!")
    return pystache.render(DICT_TEMPLATE, data)


def go(input_filename, output_filename, unit):
    xml = load_xml(input_filename)
    dct = make_dict(xml, input_filename, unit)
    save_txt(output_filename, dct)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Make a pronunciation dictionary from a G2P'd XML file"
    )
    parser.add_argument("input", type=str, help="Input XML")
    parser.add_argument("output", type=str, help="Output .dict file")
    parser.add_argument(
        "--unit",
        type=str,
        default="m",
        help="XML tag of the unit of analysis " '(e.g. "w" for word, "m" for morpheme)',
    )
    args = parser.parse_args()
    go(args.input, args.output, args.unit)



#!/usr/bin/env python3

import io
import os
from contextlib import redirect_stderr
from unittest import TestCase, main

from lxml import etree

from readalongs.text.add_elements_to_xml import add_images, add_supplementary_xml


class TestConfig(TestCase):
    @classmethod
    def setUpClass(cls):
        data_dir = os.path.join(os.path.dirname(__file__), "data")
        cls.xml = etree.parse(os.path.join(data_dir, "ej-fra.xml")).getroot()

    def test_image(self):
        """Test images are added correctly"""
        with self.assertRaises(KeyError):
            new_xml = add_images(self.xml, {})
        new_xml = add_images(self.xml, {"images": {"0": "test.jpg"}})
        self.assertTrue(len(new_xml.xpath("//graphic")) == 1)
        with self.assertRaises(TypeError):
            new_xml = add_images(self.xml, {"images": [{"0": "test.jpg"}]})
        with self.assertRaises(ValueError):
            new_xml = add_images(self.xml, {"images": {"a": "test.jpg"}})
        with self.assertRaises(IndexError):
            new_xml = add_images(
                self.xml, {"images": {"0": "test.jpg", "999": "out_of_range.jpg"}}
            )

    def test_arbitrary_xml(self):
        """Test arbitrary xml is added correctly"""
        with self.assertRaises(KeyError):
            new_xml = add_supplementary_xml(self.xml, {})
        new_xml = add_supplementary_xml(
            self.xml,
            {
                "xml": [
                    {
                        "xpath": "//div[1]",
                        "value": "<test>here is some test text</test>",
                    }
                ]
            },
        )
        self.assertTrue(len(new_xml.xpath("//test")) == 1)

        # bad xml raises lxml.etree.XMLSyntaxError
        with self.assertRaises(etree.XMLSyntaxError):
            new_xml = add_supplementary_xml(
                self.xml, {"xml": [{"xpath": "//div[1]", "value": "bloop"}]},
            )

        # if xpath isn't valid, log warning
        log_output = io.StringIO()
        with redirect_stderr(log_output):
            new_xml = add_supplementary_xml(
                self.xml,
                {
                    "xml": [
                        {
                            "xpath": "//bloop",
                            "value": "<shmoop>here is some test text</shmoop>",
                        }
                    ]
                },
            )
        self.assertIn("No elements found at //bloop", log_output.getvalue())
        self.assertTrue(len(new_xml.xpath("//shmoop")) == 0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###################################################
#
# add_elements_to_xml.py
#
# In order to streamline the image-adding process, this module
# allows graphics for pages to be specified in a supplementary json
# configuration file.
#
# It also include a more general-purpose XML markup mechanism in add_supplementary_xml
#
# See test.images for more information
#
###################################################

from lxml import etree

from readalongs.log import LOGGER


def add_images(element: etree, config: dict) -> etree:
    """Add images from configuration object to xml

    Args:
        element (etree): xml without images
        config (dict): standard ReadAlong-Studio configuration

    Returns:
        etree: xml with images markup
    """
    if "images" not in config:
        raise KeyError(
            "Configuration tried to add images, but no images were found in configuration"
        )

    if not isinstance(config["images"], dict):
        raise TypeError(
            f"Image configuration is of type {type(config['images'])} but a dict is required."
        )

    pages = element.xpath('//div[@type="page"]')

    for i, url in config["images"].items():
        image_el = etree.Element("graphic", url=url)
        try:
            i = int(i)
        except ValueError:
            raise ValueError(f"Images must be indexed using integers, you provided {i}")
        try:
            pages[int(i)].append(image_el)
        except IndexError:
            raise IndexError(
                f"No page found at index {i}, please verify your configuration"
            )

    return element


def add_supplementary_xml(element: etree, config: dict) -> etree:
    """Add arbitrary xml from configuration object to xml

    Args:
        element (etree): original xml document
        config (dict): standard ReadAlong-Studio configuration

    Returns:
        etree: xml with supplemental markup
    """
    if "xml" not in config:
        raise KeyError(
            "Configuration tried to add supplementary xml, but no declarations were found in configuration"
        )
    for el in config["xml"]:
        parents = element.xpath(el["xpath"])
        if not parents:
            LOGGER.warning(
                f"No elements found at {el['xpath']}, please verify your configuration."
            )
        for parent in parents:
            parent.append(etree.XML(el["value"]))

    return element



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

##################################################
#
# make_fsg.py
#
# This module takes a text file, marked up with
# units (e.g. w for word, m for morpheme) and ids
# and converted to IPA, and outputs a FSG
# file for processing by PocketSphinx.
#
##################################################


from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import os

import pystache
from slugify import slugify

from readalongs.log import LOGGER
from readalongs.text.util import load_xml, save_txt

FSG_TEMPLATE = """FSG_BEGIN {{name}}
NUM_STATES {{num_states}}
START_STATE 0
FINAL_STATE {{final_state}}

{{#states}}
TRANSITION {{current}} {{next}} 1.0 {{id}}
{{/states}}
FSG_END
"""


def make_fsg(xml, filename, unit="m"):
    name = slugify(os.path.splitext(os.path.basename(filename))[0])
    data = {
        "name": name,  # If name includes special characters, pocketsphinx throws a RuntimeError: new_Decoder returned -1
        "states": [],
        "num_states": 0,
    }

    for e in xml.xpath(".//" + unit):
        if "id" not in e.attrib:  # don't put in elements with no id
            continue
        if not e.text or not e.text.strip():
            LOGGER.warning("No text in node %s", e.attrib["id"])
            continue
        text = e.text.strip()
        # if not e.text.strip():  # don't put in elements with no text
        #    continue
        data["states"].append(
            {
                "id": e.attrib["id"] if text else "",
                "current": data["num_states"],
                "next": data["num_states"] + 1,
            }
        )
        data["num_states"] += 1

    data["final_state"] = data["num_states"]
    data["num_states"] += 1

    return pystache.render(FSG_TEMPLATE, data)


def go(input_filename, output_filename, unit):
    xml = load_xml(input_filename)
    fsg = make_fsg(xml, input_filename, unit)
    save_txt(output_filename, fsg)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Make an FSG grammar from an XML file with IDs"
    )
    parser.add_argument("input", type=str, help="Input XML")
    parser.add_argument("output_fsg", type=str, help="Output .fsg file")
    parser.add_argument(
        "--unit",
        type=str,
        default="m",
        help="XML tag of the unit of analysis " '(e.g. "w" for word, "m" for morpheme)',
    )
    args = parser.parse_args()
    go(args.input, args.output_fsg, args.unit)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# cli.py
#
#   Initializes a Command Line Interface with Click.
#   The main purpose of the cli is to align input files.
#
#   CLI commands implemented in this file:
#    - align  : main command to align text and audio
#    - epub   : convert aligned file to epub format
#    - prepare: prepare XML input for align from plain text
#    - tokenize: tokenize the prepared file
#    - g2p    : apply g2p to the tokenized file
#
#   Default CLI commands provided by Flask:
#    - routes : show available routes in the this readalongs Flask app
#    - run    : run the readalongs Flask app
#    - shell  : open a shell within the readalongs Flask application context
#
#######################################################################

import io
import json
import os
import sys
from tempfile import TemporaryFile

import click
from flask.cli import FlaskGroup
from lxml import etree

from readalongs._version import __version__
from readalongs.align import (
    align_audio,
    convert_to_xhtml,
    create_input_tei,
    return_words_and_sentences,
    save_readalong,
    write_to_subtitles,
    write_to_text_grid,
)
from readalongs.app import app
from readalongs.audio_utils import read_audio_from_file
from readalongs.epub.create_epub import create_epub
from readalongs.log import LOGGER
from readalongs.python_version import ensure_using_supported_python_version
from readalongs.text.add_ids_to_xml import add_ids
from readalongs.text.convert_xml import convert_xml
from readalongs.text.tokenize_xml import tokenize_xml
from readalongs.text.util import save_xml, write_xml
from readalongs.views import LANGS

ensure_using_supported_python_version()


def create_app():
    """ Returns the app
    """
    return app


def get_click_file_name(click_file):
    """ Wrapper around click_file.name with consistent handling for stdin

    On Windows, if click_file is stdin, click_file.name == "-".
    On Linux, if click_file is stdin, click_file.name == "<stdin>".
    During unit testing, the simulated stdin stream has no .name attribute

    Args:
        click_file(click.File): the click file whose name we need

    Returns:
        "-" if click_file represents stdin, click_file.name otherwise
    """
    try:
        name = click_file.name
    except Exception:
        name = "-"
    return "-" if name == "<stdin>" else name


def parse_g2p_fallback(g2p_fallback_arg):
    """ Parse the strings containing a colon-separated list of fallback args into a
        Python list of language codes, or empty if None
    """
    if g2p_fallback_arg:
        g2p_fallbacks = g2p_fallback_arg.split(":")
        for lang in g2p_fallbacks:
            if lang not in LANGS:
                raise click.BadParameter(
                    f'g2p fallback lang "{lang}" is not valid; choose among {", ".join(LANGS)}'
                )
        return g2p_fallbacks
    else:
        return []


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.version_option(version=__version__, prog_name="readalongs")
@click.group(cls=FlaskGroup, create_app=create_app, context_settings=CONTEXT_SETTINGS)
def cli():
    """Management script for Read Along Studio."""


@app.cli.command(
    context_settings=CONTEXT_SETTINGS, short_help="Force align a text and a sound file."
)
@click.argument("textfile", type=click.Path(exists=True, readable=True))
@click.argument("audiofile", type=click.Path(exists=True, readable=True))
@click.argument("output-base", type=click.Path())
@click.option(
    "-b",
    "--bare",
    is_flag=True,
    help="Bare alignments do not split silences between words",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True),
    help="Use ReadAlong-Studio configuration file (in JSON format)",
)
@click.option(
    "-C",
    "--closed-captioning",
    is_flag=True,
    help="Export sentences to WebVTT and SRT files",
)
@click.option("-d", "--debug", is_flag=True, help="Add debugging messages to logger")
@click.option(
    "-f", "--force-overwrite", is_flag=True, help="Force overwrite output files"
)
@click.option(
    "-i",
    "--text-input",
    is_flag=True,
    help="Input is plain text (assume paragraphs are separated by blank lines, pages are separated by two blank lines)",
)
@click.option(
    "-l",
    "--language",
    type=click.Choice(LANGS, case_sensitive=False),
    help="Set language for plain text input",
)
@click.option(
    "-u",
    "--unit",
    type=click.Choice(["w", "m"], case_sensitive=False),
    help="Unit (w = word, m = morpheme) to align to",
)
@click.option(
    "-s",
    "--save-temps",
    is_flag=True,
    help="Save intermediate stages of processing and temporary files (dictionary, FSG, tokenization etc)",
)
@click.option(
    "-t", "--text-grid", is_flag=True, help="Export to Praat TextGrid & ELAN eaf file"
)
@click.option(
    "-x", "--output-xhtml", is_flag=True, help="Output simple XHTML instead of XML"
)
@click.option(
    "--g2p-fallback",
    default=None,
    help="Colon-separated list of fallback langs for g2p; enables the g2p cascade",
)
@click.option(
    "--g2p-verbose",
    is_flag=True,
    default=False,
    help="Display verbose messages about g2p errors.",
)
def align(**kwargs):  # noqa: C901
    """Align TEXTFILE and AUDIOFILE and create output files as OUTPUT_BASE.* in directory
    OUTPUT_BASE/.

    TEXTFILE:    Input text file path (in XML, or plain text with -i)

    With -i, TEXTFILE should be plain text;
    without -i, TEXTFILE can be in one of three XML formats:
    the output of 'readalongs prepare',
    the output of 'readalongs tokenize', or
    the output of 'readalongs g2p'.

    One can also add the known ARPABET encoding for words (<w> elements) that
    are not correctly handled by g2p in the output of 'readalongs tokenize' or
    'readalongs g2p', via the ARPABET attribute.

    AUDIOFILE:   Input audio file path, in any format supported by ffmpeg

    OUTPUT_BASE: Base name for output files
    """
    config_file = kwargs.get("config", None)
    config = None
    if config_file:
        if config_file.endswith("json"):
            try:
                with open(config_file) as f:
                    config = json.load(f)
            except json.decoder.JSONDecodeError:
                raise click.BadParameter(
                    f"Config file at {config_file} is not in valid JSON format."
                )
        else:
            raise click.BadParameter(
                f"Config file '{config_file}' must be in JSON format"
            )

    output_dir = kwargs["output_base"]
    if os.path.exists(output_dir):
        if not os.path.isdir(output_dir):
            raise click.UsageError(
                f"Output folder '{output_dir}' already exists but is a not a directory."
            )
        if not kwargs["force_overwrite"]:
            raise click.UsageError(
                f"Output folder '{output_dir}' already exists, use -f to overwrite."
            )
    else:
        os.mkdir(output_dir)

    # Make sure we can write to the output directory, for early error checking and user
    # friendly error messages.
    try:
        with TemporaryFile(dir=output_dir):
            pass
    except Exception:
        raise click.UsageError(
            f"Cannot write into output folder '{output_dir}'. Please verify permissions."
        )

    output_basename = os.path.basename(output_dir)
    temp_base = None
    if kwargs["save_temps"]:
        temp_dir = os.path.join(output_dir, "tempfiles")
        if not os.path.isdir(temp_dir):
            if os.path.exists(temp_dir) and kwargs["force_overwrite"]:
                os.unlink(temp_dir)
            os.mkdir(temp_dir)
        temp_base = os.path.join(temp_dir, output_basename)

    if kwargs["debug"]:
        LOGGER.setLevel("DEBUG")

    if kwargs["text_input"]:
        if not kwargs["language"]:
            LOGGER.warning("No input language provided, using undetermined mapping")
        plain_textfile = kwargs["textfile"]
        _, xml_textfile = create_input_tei(
            input_file_name=plain_textfile,
            text_language=kwargs["language"],
            save_temps=temp_base,
        )
    else:
        xml_textfile = kwargs["textfile"]

    unit = kwargs.get("unit", "w") or "w"  # Sometimes .get() still returns None here
    bare = kwargs.get("bare", False)

    try:
        results = align_audio(
            xml_textfile,
            kwargs["audiofile"],
            unit=unit,
            bare=bare,
            config=config,
            save_temps=temp_base,
            g2p_fallbacks=parse_g2p_fallback(kwargs["g2p_fallback"]),
            verbose_g2p_warnings=kwargs["g2p_verbose"],
        )
    except RuntimeError as e:
        LOGGER.error(e)
        exit(1)

    save_readalong(
        align_results=results,
        output_dir=output_dir,
        output_basename=output_basename,
        config=config,
        text_grid=kwargs["text_grid"],
        closed_captioning=kwargs["closed_captioning"],
        output_xhtml=kwargs["output_xhtml"],
        audiofile=kwargs["audiofile"],
    )


@app.cli.command(
    context_settings=CONTEXT_SETTINGS, short_help="Convert a smil document to epub."
)
@click.argument("input", type=click.Path(exists=True, readable=True))
@click.argument("output", type=click.Path(exists=False, readable=True))
@click.option(
    "-u",
    "--unpacked",
    is_flag=True,
    help="Output unpacked directory of files (for testing)",
)
def epub(**kwargs):
    """
    Convert INPUT smil document to epub with media overlay at OUTPUT

    INPUT:  The .smil document

    OUTPUT: Path to the .epub output
    """
    create_epub(kwargs["input"], kwargs["output"], kwargs["unpacked"])


@app.cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Convert a plain text file into the XML format for alignment.",
)
@click.argument("plaintextfile", type=click.File("r", encoding="utf8", lazy=True))
@click.argument("xmlfile", type=click.Path(), required=False, default="")
@click.option("-d", "--debug", is_flag=True, help="Add debugging messages to logger")
@click.option(
    "-f", "--force-overwrite", is_flag=True, help="Force overwrite output files"
)
@click.option(
    "-l",
    "--language",
    type=click.Choice(LANGS, case_sensitive=False),
    required=True,
    help="Set language for input file",
)
def prepare(**kwargs):
    """Prepare XMLFILE for 'readalongs align' from PLAINTEXTFILE.
    PLAINTEXTFILE must be plain text encoded in utf-8, with one sentence per line,
    paragraph breaks marked by a blank line, and page breaks marked by two
    blank lines.

    PLAINTEXTFILE: Path to the plain text input file, or - for stdin

    XMLFILE:       Path to the XML output file, or - for stdout [default: PLAINTEXTFILE.xml]
    """

    if kwargs["debug"]:
        LOGGER.setLevel("DEBUG")
        LOGGER.info(
            "Running readalongs prepare(lang={}, force-overwrite={}, plaintextfile={}, xmlfile={}).".format(
                kwargs["language"],
                kwargs["force_overwrite"],
                kwargs["plaintextfile"],
                kwargs["xmlfile"],
            )
        )

    input_file = kwargs["plaintextfile"]

    out_file = kwargs["xmlfile"]
    if not out_file:
        out_file = get_click_file_name(input_file)
        if out_file != "-":
            if out_file.endswith(".txt"):
                out_file = out_file[:-4]
            out_file += ".xml"

    if out_file == "-":
        _, filename = create_input_tei(
            input_file_handle=input_file, text_language=kwargs["language"],
        )
        with io.open(filename, encoding="utf8") as f:
            sys.stdout.write(f.read())
    else:
        if not out_file.endswith(".xml"):
            out_file += ".xml"
        if os.path.exists(out_file) and not kwargs["force_overwrite"]:
            raise click.BadParameter(
                "Output file %s exists already, use -f to overwrite." % out_file
            )

        _, filename = create_input_tei(
            input_file_handle=input_file,
            text_language=kwargs["language"],
            output_file=out_file,
        )

    LOGGER.info("Wrote {}".format(out_file))


@app.cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Tokenize a prepared XML file, in preparation for alignment.",
)
@click.argument("xmlfile", type=click.File("rb"))
@click.argument("tokfile", type=click.Path(), required=False, default="")
@click.option("-d", "--debug", is_flag=True, help="Add debugging messages to logger")
@click.option(
    "-f", "--force-overwrite", is_flag=True, help="Force overwrite output files"
)
def tokenize(**kwargs):
    """Tokenize XMLFILE for 'readalongs align' into TOKFILE.
    XMLFILE should have been produced by 'readalongs prepare'.
    TOKFILE can then be augmented with word-specific language codes.
    'readalongs align' can be called with either XMLFILE or TOKFILE as XML input.

    XMLFILE: Path to the XML file to tokenize, or - for stdin

    TOKFILE: Output path for the tok'd XML, or - for stdout [default: XMLFILE.tokenized.xml]
    """

    if kwargs["debug"]:
        LOGGER.setLevel("DEBUG")
        LOGGER.info(
            "Running readalongs tokenize(xmlfile={}, tokfile={}, force-overwrite={}).".format(
                kwargs["xmlfile"], kwargs["tokfile"], kwargs["force_overwrite"],
            )
        )

    input_file = kwargs["xmlfile"]

    if not kwargs["tokfile"]:
        output_path = get_click_file_name(input_file)
        if output_path != "-":
            if output_path.endswith(".xml"):
                output_path = output_path[:-4]
            output_path += ".tokenized.xml"
    else:
        output_path = kwargs["tokfile"]
        if not output_path.endswith(".xml") and not output_path == "-":
            output_path += ".xml"

    if os.path.exists(output_path) and not kwargs["force_overwrite"]:
        raise click.BadParameter(
            "Output file %s exists already, use -f to overwrite." % output_path
        )

    try:
        xml = etree.parse(input_file).getroot()
    except etree.XMLSyntaxError as e:
        raise click.BadParameter(
            "Error parsing input file %s as XML, please verify it. Parser error: %s"
            % (get_click_file_name(input_file), e)
        )

    # Tokenize the XML file - all this code for such a tiny body!!!
    xml = tokenize_xml(xml)

    if output_path == "-":
        write_xml(sys.stdout.buffer, xml)
    else:
        save_xml(output_path, xml)
    LOGGER.info("Wrote {}".format(output_path))


@app.cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Apply g2p to a tokenized file, like 'align' does.",
    # NOT TRUE YET: "Apply g2p to a tokenized file, in preparation for alignment."
)
@click.argument("tokfile", type=click.File("rb", encoding="utf8", lazy=True))
@click.argument("g2pfile", type=click.Path(), required=False, default="")
@click.option(
    "--g2p-fallback",
    default=None,
    help="Colon-separated list of fallback langs for g2p; enables the g2p cascade",
)
@click.option(
    "-f", "--force-overwrite", is_flag=True, help="Force overwrite output files"
)
@click.option(
    "--g2p-verbose",
    is_flag=True,
    default=False,
    help="Display verbose messages about g2p errors.",
)
@click.option("-d", "--debug", is_flag=True, help="Add debugging messages to logger")
def g2p(**kwargs):
    """Apply g2p mappings to TOKFILE into G2PFILE.
    TOKFILE should have been produced by 'readalongs tokenize'.
    G2PFILE can then be modified to adjust the phonetic representation as needed.
    'readalongs align' can be called with G2PFILE instead of TOKFILE as XML input.

    WARNING: the output is not yet compatible with align and cannot be used as input to align.

    TOKFILE: Path to the input tokenized XML file, or - for stdin

    G2PFILE: Output path for the g2p'd XML, or - for stdout [default: TOKFILE with .g2p. inserted]
    """
    if kwargs["debug"]:
        LOGGER.setLevel("DEBUG")
        LOGGER.info(
            "Running readalongs g2p(tokfile={}, g2pfile={}, force-overwrite={}).".format(
                kwargs["tokfile"], kwargs["g2pfile"], kwargs["force_overwrite"],
            )
        )

    input_file = kwargs["tokfile"]

    if not kwargs["g2pfile"]:
        output_path = get_click_file_name(input_file)
        if output_path != "-":
            if output_path.endswith(".xml"):
                output_path = output_path[:-4]
            if output_path.endswith(".tokenized"):
                output_path = output_path[: -len(".tokenized")]
            output_path += ".g2p.xml"
    else:
        output_path = kwargs["g2pfile"]
        if not output_path.endswith(".xml") and not output_path == "-":
            output_path += ".xml"

    if os.path.exists(output_path) and not kwargs["force_overwrite"]:
        raise click.BadParameter(
            "Output file %s exists already, use -f to overwrite." % output_path
        )

    try:
        xml = etree.parse(input_file).getroot()
    except etree.XMLSyntaxError as e:
        raise click.BadParameter(
            "Error parsing input file %s as XML, please verify it. Parser error: %s"
            % (get_click_file_name(input_file), e)
        )

    # Add the IDs to paragraph, sentences, word, etc.
    xml = add_ids(xml)
    # Apply the g2p mappings.
    xml, valid = convert_xml(
        xml,
        g2p_fallbacks=parse_g2p_fallback(kwargs["g2p_fallback"]),
        verbose_warnings=kwargs["g2p_verbose"],
    )

    if output_path == "-":
        write_xml(sys.stdout.buffer, xml)
    else:
        save_xml(output_path, xml)
        LOGGER.info("Wrote {}".format(output_path))

    if not valid:
        LOGGER.error(
            "Some word(s) could not be g2p'd correctly."
            + (
                " Run again with --g2p-verbose to get more detailed error messages."
                if not kwargs["g2p_verbose"]
                else ""
            )
        )
        exit(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###################################################
#
# waveform2svg.py
#
# Make an SVG of a waveform, using roughly the same
# procedure as Audacity uses: chop up the samples into
# n buckets, plot only the maximum and minimum values.
#
###################################################

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import os
from io import open
from math import ceil, floor

import numpy as np
import pystache

from readalongs.log import LOGGER
from readalongs.waveform2svg.audio_util import load_wav_or_smil, save_txt, smooth

SVG_TEMPLATE = """<svg id='wave' preserveAspectRatio='none' viewBox="0 0 {{width}} {{height}}" xmlns="http://www.w3.org/2000/svg" height="{{height}}" width="{{width}}">
  <defs>
      <linearGradient id="Gradient" x1="0" x2="1" y1="0" y2="0">
        <stop id='progress-base' offset="0%"/>
        <stop id='progress-fill' class='' offset="0%"/>
        <stop id='progress-trail' offset='0%' stop-color='transparent'/>
      </linearGradient>
  </defs>
    <polygon id="polygon" stroked-width=".65" stroke="black" fill="url(#Gradient)" points="{{#points}}{{x}},{{y}} {{/points}}"></polygon>
</svg>
"""

SAMPLE_RATE = 16000


def get_max_and_min(data, nbuckets):
    data /= np.abs(data).max(axis=0)  # normalize
    num_samples = data.shape[0]
    samples_per_bucket = int(ceil(num_samples / nbuckets))
    max_amps, min_amps = [], []
    for i in range(0, num_samples, samples_per_bucket):
        max_amps.append(data[i : i + 256].max())
        min_amps.append(data[i : i + 256].min())
    max_amps = list(smooth(max_amps, window_size=int(floor(nbuckets / 100))))
    min_amps = list(smooth(min_amps, window_size=int(floor(nbuckets / 100))))
    return max_amps, min_amps


def render_svg(
    max_amps,
    min_amps,
    num_buckets,
    include_negative=True,
    width=512,
    height=100,
    zero_height=5,
):
    data = {"height": height, "width": width, "points": []}
    data["points"].append({"x": 0.0, "y": height / 2})
    data["points"].append({"x": 0.0, "y": (height - zero_height) / 2})
    for i, max_amp in enumerate(max_amps):
        y = (1.0 - max_amp) * ((height - zero_height) / 2)
        y = "%.2f" % y
        x = i / num_buckets * width + 0.5
        data["points"].append({"x": x, "y": y})
    data["points"].append({"x": float(width), "y": (height - zero_height) / 2})
    data["points"].append({"x": float(width), "y": height / 2})
    if include_negative:
        data["points"].append({"x": float(width), "y": (height + zero_height) / 2})
        min_amps.reverse()
        for i, min_amp in enumerate(min_amps):
            y = (1.0 - min_amp) * ((height + zero_height) / 2)
            y = "%.2f" % y
            x = width - (i / num_buckets * width + 0.5)
            data["points"].append({"x": x, "y": y})
        data["points"].append({"x": 0.0, "y": (height + zero_height) / 2})
    return pystache.render(SVG_TEMPLATE, data)


def make_waveform_svg(
    input_path, num_buckets=512, include_neg=True, width=512, height=100, zero_height=5
):
    waveform = load_wav_or_smil(input_path)
    max_amps, min_amps = get_max_and_min(waveform, num_buckets)
    return render_svg(
        max_amps, min_amps, num_buckets, include_neg, width, height, zero_height
    )


def main(
    input_path,
    output_path,
    num_buckets=512,
    include_neg=True,
    width=512,
    height=100,
    zero_height=5,
):
    svg = make_waveform_svg(
        input_path, num_buckets, include_neg, width, height, zero_height
    )
    save_txt(output_path, svg)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert a WAV file to a SVG file of its waveform"
    )
    parser.add_argument("input", type=str, help="Input WAV file")
    parser.add_argument("output", type=str, help="Output SVG file")
    parser.add_argument(
        "--nbuckets",
        type=int,
        default=512,
        help="Number of sample buckets (default: 256)",
    )
    parser.add_argument(
        "--include_neg",
        type=bool,
        default=True,
        help="Include negative values? (default: True",
    )
    parser.add_argument(
        "--width", type=int, default=512, help="Width of output SVG (default: 512)"
    )
    parser.add_argument(
        "--height", type=int, default=100, help="Height of output SVG (default: 100)"
    )
    parser.add_argument(
        "--zero_height", type=int, default=5, help="Padding around zero (default: 5)"
    )
    args = parser.parse_args()
    main(
        args.input,
        args.output,
        args.nbuckets,
        args.include_neg,
        args.width,
        args.height,
        args.zero_height,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###################################################
#
# pitch2svg.py
#
# Make an SVG of a pitch trace.
#
# Based loosely off code from Martijn Millecamp
# (martijn.millecamp@student.kuleuven.be) and
# Miroslav Masat (miro.masat@gmail.com):
#  https://github.com/miromasat/pitch-detection-librosa-python
#
###################################################

from __future__ import absolute_import, division, print_function, unicode_literals

import os
from io import open

import librosa
import numpy as np
from lxml import etree

from readalongs.text.util import xpath_default

SAMPLE_RATE = 16000


def smooth(x, window_size=5):
    """ Smooth the waveform to look... well, smooth """
    if window_size < 3:
        return x
    s = np.r_[2 * x[0] - x[window_size - 1 :: -1], x, 2 * x[-1] - x[-1:-window_size:-1]]
    w = np.hanning(window_size)
    y = np.convolve(w / w.sum(), s, mode="same")
    return y[window_size : -window_size + 1]


def load_smil(input_path):
    """ Get the bucketed max and min value from a sequence of WAV files as
        expressed in a SMIL document """
    xml = load_xml(input_path)
    dirname = os.path.dirname(input_path)
    data = None
    most_recent_audio_src = None
    for audio_node in xpath_default(xml, ".//i:audio"):
        audio_src = audio_node.attrib["src"]
        if audio_src == most_recent_audio_src:
            continue
        most_recent_audio_src = audio_src
        audio_path = os.path.join(dirname, audio_src)
        waveform = load_wav(audio_path)
        data = np.hstack((data, waveform)) if data is not None else waveform
    return data


def load_wav(input_path):
    waveform, _ = librosa.load(input_path, sr=SAMPLE_RATE)
    return waveform


def load_wav_or_smil(input_path):
    if os.path.splitext(input_path)[1].lower() == ".smil":
        waveform = load_smil(input_path)
    else:
        waveform = load_wav(input_path)
    return waveform


def ensure_dirs(path):
    dirname = os.path.dirname(path)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)


def load_xml(input_path):
    with open(input_path, "r", encoding="utf-8") as fin:
        return etree.fromstring(fin.read())


def save_txt(output_path, txt):
    ensure_dirs(output_path)
    with open(output_path, "w", encoding="utf-8") as fout:
        fout.write(txt)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# audio_utils.py
#
#   Various utility functions for manipulating audio
#
#######################################################################

from pydub import AudioSegment

from readalongs.log import LOGGER


def join_section(audio: AudioSegment, audio_to_insert: AudioSegment, start: int):
    """ Given two AudioSegments, insert the second into the first at start (ms)
    """
    try:
        return audio[:start] + audio_to_insert + audio[start:]
    except IndexError:
        LOGGER.error(
            f"Tried to insert audio at {start}, but audio is only {len(audio)}ms long. \
                     Returning unchanged audio instead."
        )
        return audio


def remove_section(audio: AudioSegment, start: int, end: int) -> AudioSegment:
    """ Given an AudioSement, remove the section between start (ms) and end (ms)
    """
    try:
        return audio[:start] + audio[end:]
    except IndexError:
        LOGGER.error(
            f"Tried to remove audio between {start} and {end}, but audio is only {len(audio)}ms long. \
                     Returning unchanged audio instead."
        )
        return audio


def mute_section(audio: AudioSegment, start: int, end: int) -> AudioSegment:
    """ Given an AudioSegment, reduce the gain between a given interval by 120db.
        Effectively, make it silent.

    Parameters
    ----------
    audio : AudioSegment
        audio segment to mute
    start : int
        start timestamp of audio (ms)
    end : int
        end timestamp of audio (ms)

    Returns
    -------
    AudioSegment
        A muted audio segment
    """
    try:
        return audio[:start] + audio[start:end].apply_gain(-120) + audio[end:]
    except IndexError:
        LOGGER.error(
            f"Tried to mute audio between {start} and {end}, but audio is only {len(audio)}ms long. \
                     Returning unmuted audio instead."
        )
        return audio


def write_audio_to_file(audio: AudioSegment, path: str) -> None:
    """ Write AudioSegment to file
        TODO: Add params/file type kwargs
        TODO: Add exception handling
    """
    audio.export(path)


def read_audio_from_file(path: str) -> AudioSegment:
    """ Read in AudioSegment from file

    Parameters
    ----------
    str
        Path to audiofile

    Returns
    -------
    AudioSegment
        An AudioSegment object of the audiofile

    Raises
    ------
    RuntimeError
        catches empty audio files and other problems with them.
    """
    try:
        return AudioSegment.from_file(path)
    except Exception as e:
        # need repr(e) here instead of e since these exceptions don't all have messages
        raise RuntimeError("Error reading audio file %s: %s" % (path, repr(e)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# log.py
#
#   Setup a logger that has colours!
#
#######################################################################

import logging

import coloredlogs

FIELD_STYLES = dict(levelname=dict(color="green", bold=coloredlogs.CAN_USE_BOLD_FONT),)


def setup_logger(name):
    """ Create logger and configure with cool colors!
    """
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(name)

    # Use level='NOTSET' (most permissive) here so whatever level the user later selects
    # does get printed. with level='INFO' here, setting LOGGER.setLevel('DEBUG') in the
    # app doesn't work, and therefore the --debug command line options doesn't work.
    coloredlogs.install(
        level="NOTSET",
        fmt="%(levelname)s - %(message)s",
        logger=logger,
        field_styles=FIELD_STYLES,
    )

    logger.setLevel("INFO")  # default logging level is INFO
    return logger


LOGGER = setup_logger("root")



#!/usr/bin/env python3

"""
Test PortableNamedTemporaryFile class
"""

import os
import sys
import unittest
from tempfile import NamedTemporaryFile

from readalongs.log import LOGGER
from readalongs.portable_tempfile import PortableNamedTemporaryFile


class TestTempFile(unittest.TestCase):
    def testBasicFile(self):
        f = open("delme_test_temp_file", mode="w")
        f.write("some text")
        f.close()
        os.unlink("delme_test_temp_file")

    def testNTF(self):
        tf = NamedTemporaryFile(prefix="testtempfile_testNTF_", delete=False, mode="w")
        tf.write("Some text")
        # LOGGER.debug("tf.name {}".format(tf.name))
        tf.close()
        readf = open(tf.name, mode="r")
        text = readf.readline()
        self.assertEqual(text, "Some text")
        readf.close()
        os.unlink(tf.name)

    def testDeleteFalse(self):
        tf = PortableNamedTemporaryFile(
            prefix="testtempfile_testDeleteFalse_", delete=False, mode="w"
        )
        tf.write("Some text")
        tf.close()
        # LOGGER.info(tf.name)
        readf = open(tf.name, mode="r")
        text = readf.readline()
        readf.close()
        self.assertEqual(text, "Some text")
        os.unlink(tf.name)

    def testTypicalUsage(self):
        tf = PortableNamedTemporaryFile(
            prefix="testtempfile_testTypicalUsage_", delete=True, mode="w"
        )
        # LOGGER.info(tf.name)
        tf.write("Some text")
        tf.close()
        # LOGGER.info(tf.name)
        readf = open(tf.name, mode="r")
        text = readf.readline()
        readf.close()
        self.assertEqual(text, "Some text")

    def testUsingWith(self):
        with PortableNamedTemporaryFile(
            prefix="testtempfile_testUsingWith_", delete=True, mode="w"
        ) as tf:
            # LOGGER.info(tf.name)
            tf.write("Some text")
            tf.close()
            # LOGGER.info(tf.name)
            readf = open(tf.name, mode="r")
            text = readf.readline()
            readf.close()
            self.assertEqual(text, "Some text")

    def testSeek(self):
        tf = PortableNamedTemporaryFile(
            prefix="testtempfile_testSeek_", delete=True, mode="w+"
        )
        tf.write("Some text")
        tf.seek(0)
        text = tf.readline()
        self.assertEqual(text, "Some text")
        tf.close()
        os.unlink(tf.named_temporary_file.name)


if __name__ == "__main__":
    LOGGER.setLevel("DEBUG")
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#########################################################################
#
# end_to_end.py
#
# Takes an XML file (preferrably using TEI conventions) and
# makes:
#
# 1. An XML file with added IDs for elements (if the elements didn't
#    already have ID attributes)
# 2. An FSG file where the transitions are those IDs.
# 3. A dictionary file giving a mapping between IDs and approximate
#    pronunciations in ARPABET
#
#
# The XML file needs to have xml:lang attributes; if tokenization is
# not performed (indicated with <w> elements) it will be attempted
# automatically.  Alignment can be done at any level of analysis; if
# there are, for example, morpheme tags (<m>), you can make that be
# the level of analysis with the option --unit m
#
# TODO: Add numpy standard docstrings to functions
##########################################################################

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse

from readalongs.log import LOGGER
from readalongs.text.add_ids_to_xml import add_ids
from readalongs.text.convert_xml import convert_xml
from readalongs.text.make_dict import make_dict
from readalongs.text.make_fsg import make_fsg
from readalongs.text.tokenize_xml import tokenize_xml
from readalongs.text.util import load_xml, save_txt, save_xml


def end_to_end(xml, input_filename, unit, word_unit, out_orth):
    xml = tokenize_xml(xml)
    xml = add_ids(xml)
    converted_xml, valid = convert_xml(xml, word_unit, out_orth)
    # save_xml("test.xml", converted_xml)
    fsg = make_fsg(converted_xml, input_filename, unit)
    pronouncing_dictionary = make_dict(converted_xml, input_filename, unit)
    return xml, fsg, pronouncing_dictionary


def go(
    input_filename,
    output_xml_filename,
    output_fsg_filename,
    output_dict_filename,
    unit,
    word_unit,
    out_orth,
):
    xml = load_xml(input_filename)
    xml, fsg, dct = end_to_end(xml, input_filename, unit, word_unit, out_orth)
    save_xml(output_xml_filename, xml)
    save_txt(output_fsg_filename, fsg)
    save_txt(output_dict_filename, dct)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert XML to another orthography while preserving tags"
    )
    parser.add_argument("input", type=str, help="Input XML")
    parser.add_argument("output_xml", type=str, help="Output XML file")
    parser.add_argument("output_fsg", type=str, help="Output .jsgf file")
    parser.add_argument("output_dict", type=str, help="Output .dict file")
    parser.add_argument(
        "--unit",
        type=str,
        default="w",
        help="XML tag of the unit of analysis " '(e.g. "w" for word, "m" for morpheme)',
    )
    parser.add_argument(
        "--word_unit",
        type=str,
        default="w",
        help='XML element that represents a word (default: "w")',
    )
    parser.add_argument(
        "--out_orth",
        type=str,
        default="eng-arpabet",
        help='Output orthography (default: "eng-arpabet")',
    )
    parser.add_argument("--debug", action="store_true", help="Enable debugging output")
    args = parser.parse_args()
    if args.debug:
        LOGGER.setLevel("DEBUG")
    go(
        args.input,
        args.output_xml,
        args.output_fsg,
        args.output_dict,
        args.unit,
        args.word_unit,
        args.out_orth,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

##################################################
#
# tokenize_xml.py
#
# The XML G2P conversion module expects XML with word
# annotations (specifically, <w> elements from the
# TEI conventions).  If this doesn't exist, this module
# attempts to make them.  It will assume alphanumeric characters
# are parts of words, and also any characters that it finds in the
# language's inventory (as indicated by xml:lang attributes on elements).
#
# The nice part about taking into account xml:lang tags and a
# language's inventory is that it can catch things like
# Mohawk ":" as being an alphabetic character after particular letters
# within Mohawk words, but not count colons as alphabetic characters
# in other circumstances (English words, after sounds that can't be
# geminated, etc.)
#
# Note that if you have *subword* annotations already, but
# not word annotations, this module will probably not do
# what you want; it will put the word annotations inside the
# subword annotations!  If you have subword annotations already,
# add the word annotations however is appropriate for your
# formatting conventions; the possibilities are too open ended
# for this module to attempt to guess for you.
#
##################################################


from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import re
from copy import deepcopy

from g2p.mappings import Mapping
from g2p.mappings.langs import MAPPINGS_AVAILABLE
from g2p.mappings.tokenizer import get_tokenizer
from lxml import etree

from readalongs.log import LOGGER
from readalongs.text.util import (
    get_lang_attrib,
    is_do_not_align,
    load_xml,
    save_xml,
    unicode_normalize_xml,
)


class XMLTokenizer:
    def add_word_children(self, element):
        tag = etree.QName(element.tag).localname
        nsmap = element.nsmap if hasattr(element, "nsmap") else element.getroot().nsmap
        if tag in ["w", "teiHeader", "head"]:  # don't do anything to existing words!
            new_element = deepcopy(element)
            new_element.tail = ""  # just take off their .tail so that it's not doubled
            return new_element  # as the calling method tends to it

        if is_do_not_align(element):  # skip elements marked do-not-align="true"
            new_element = deepcopy(element)
            new_element.tail = ""  # don't add spurious whitespace
            return new_element

        new_element = etree.Element(element.tag, nsmap=nsmap)
        for key, value in element.attrib.items():
            new_element.attrib[key] = value

        lang = get_lang_attrib(element)
        tokenizer = get_tokenizer(lang)
        if element.text:
            new_element.text = ""
            for unit in tokenizer.tokenize_text(element.text):
                if unit["is_word"]:
                    new_child_element = etree.Element("w", nsmap=nsmap)
                    new_child_element.text = unit["text"]
                    new_element.append(new_child_element)
                    continue
                if new_element.getchildren():
                    if not new_element[-1].tail:
                        new_element[-1].tail = ""
                    new_element[-1].tail += unit["text"]
                    continue
                new_element.text += unit["text"]

        for child in element:
            # Comments Cause Crashes so Copy them Cleanly
            if child.tag is etree.Comment:
                new_element.append(child)
                continue
            new_child_element = self.add_word_children(child)
            new_element.append(new_child_element)
            if child.tail:
                # new_element.tail = ''  # in case it's a copy
                for unit in tokenizer.tokenize_text(child.tail):
                    if unit["is_word"]:
                        new_child_element = etree.Element("w")
                        new_child_element.text = unit["text"]
                        new_element.append(new_child_element)
                        continue
                    if not new_element[-1].tail:
                        new_element[-1].tail = ""
                    new_element[-1].tail += unit["text"]

        return new_element


def tokenize_xml(xml):
    tokenizer = XMLTokenizer()
    xml = deepcopy(xml)
    # FIXME: different langs have different normalizations, is this necessary?
    unicode_normalize_xml(xml)
    words = xml.xpath(".//w")
    if words:
        LOGGER.info("Words (<w>) already present; skipping tokenization")
        return xml
    LOGGER.info("Words (<w>) not present; tokenizing")
    return tokenizer.add_word_children(xml)


def go(input_filename, output_filename):
    xml = load_xml(input_filename)
    xml = tokenize_xml(xml)
    save_xml(output_filename, xml)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert XML to another orthography while preserving tags"
    )
    parser.add_argument("input", type=str, help="Input XML")
    parser.add_argument("output", type=str, help="Output XML")
    args = parser.parse_args()
    go(args.input, args.output)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

###########################################
#
# util.py
#
# Just some shared functions
#
# TODO: Add numpy standard format docstrings
############################################

from __future__ import division, print_function, unicode_literals

import json
import os
import zipfile
from collections import OrderedDict
from copy import deepcopy
from io import TextIOWrapper, open
from unicodedata import category, normalize

from lxml import etree

from readalongs.log import LOGGER

# removed "try: unicode() except" block (was for Python 2), but this file uses unicode()
# too many times, so define it anyway.
unicode = str


def ensure_dirs(path):
    dirname = os.path.dirname(path)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)


def xpath_default(xml, query, default_namespace_prefix="i"):
    nsmap = xml.nsmap if hasattr(xml, "nsmap") else xml.getroot().nsmap
    nsmap = dict(
        ((x, y) if x else (default_namespace_prefix, y)) for (x, y) in nsmap.items()
    )
    for e in xml.xpath(query, namespaces=nsmap):
        yield e


def get_lang_attrib(element):
    lang_path = element.xpath("./@xml:lang")
    if not lang_path and "lang" in element.attrib:
        lang_path = element.attrib["lang"]
    if not lang_path and element.getparent() is not None:
        return get_lang_attrib(element.getparent())
    if not lang_path:
        return None
    return lang_path[0]


def is_do_not_align(element):
    dna = element.attrib.get("do-not-align", "")
    return dna == "true" or dna == "True" or dna == "TRUE" or dna == "1"


def load_xml(input_path):
    with open(input_path, "rb") as fin:
        return etree.fromstring(fin.read())


def load_xml_zip(zip_path, input_path):
    with zipfile.ZipFile(zip_path, "r") as fin_zip:
        with fin_zip.open(input_path, "r") as fin:
            return etree.fromstring(fin)


def load_xml_with_encoding(input_path):
    """ etree.fromstring messes up on declared encodings """
    return etree.parse(input_path)


def write_xml(output_filelike, xml):
    """ Write XML to already opened file-like object """
    output_filelike.write(etree.tostring(xml, encoding="utf-8", xml_declaration=True))
    output_filelike.write("\n".encode("utf-8"))


def save_xml(output_path, xml):
    """ Save XML to specific PATH """
    ensure_dirs(output_path)
    with open(output_path, "wb") as fout:
        write_xml(fout, xml)


def save_xml_zip(zip_path, output_path, xml):
    ensure_dirs(zip_path)
    with zipfile.ZipFile(zip_path, "a", compression=zipfile.ZIP_DEFLATED) as fout_zip:
        fout_zip.writestr(
            output_path,
            etree.tostring(xml, encoding="utf-8", xml_declaration=True) + "\n",
        )


def load_txt(input_path):
    with open(input_path, "r", encoding="utf-8") as fin:
        return fin.read()


def load_txt_zip(zip_path, input_path):
    with zipfile.ZipFile(zip_path, "r") as fin_zip:
        with fin_zip.open(input_path, "r") as fin:
            fin_utf8 = TextIOWrapper(fin, encoding="utf-8")
            return fin_utf8.read()


def save_txt(output_path, txt):
    ensure_dirs(output_path)
    with open(output_path, "w", encoding="utf-8") as fout:
        fout.write(txt)


def save_txt_zip(zip_path, output_path, txt):
    ensure_dirs(zip_path)
    with zipfile.ZipFile(zip_path, "a", compression=zipfile.ZIP_DEFLATED) as fout_zip:
        fout_zip.writestr(output_path, txt.encode("utf-8"))


def load_json(input_path):
    with open(input_path, "r", encoding="utf-8") as fin:
        return json.load(fin, object_pairs_hook=OrderedDict)


def load_json_zip(zip_path, input_path):
    with zipfile.ZipFile(zip_path, "r") as fin_zip:
        with fin_zip.open(input_path, "r") as fin:
            fin_utf8 = TextIOWrapper(fin, encoding="utf-8")
            return json.loads(fin_utf8.read(), object_pairs_hook=OrderedDict)


def save_json(output_path, obj):
    ensure_dirs(output_path)
    with open(output_path, "w", encoding="utf-8") as fout:
        fout.write(unicode(json.dumps(obj, ensure_ascii=False, indent=4)))


def save_json_zip(zip_path, output_path, obj):
    ensure_dirs(zip_path)
    txt = unicode(json.dumps(obj, ensure_ascii=False, indent=4))
    with zipfile.ZipFile(zip_path, "a") as fout_zip:
        fout_zip.writestr(output_path, txt.encode("utf-8"))


def copy_file_to_zip(zip_path, origin_path, destination_path):
    ensure_dirs(zip_path)
    with zipfile.ZipFile(zip_path, "a", compression=zipfile.ZIP_DEFLATED) as fout_zip:
        fout_zip.write(origin_path, destination_path)


MINIMAL_INDEX_HTML_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="UTF-8">
        <title>Insert Title Here</title>
        <!-- Import fonts. Material Icons are needed by the web component -->
        <link href="https://fonts.googleapis.com/css?family=Lato%7CMaterial+Icons%7CMaterial+Icons+Outlined" rel="stylesheet">
    </head>

    <body>
        <!-- Here is how you declare the Web Component. Supported languages: en, fr -->
        <read-along text="{}" alignment="{}" audio="{}" language="en">
            <span slot="read-along-header">Insert Title Here Too</span>
        </read-along>
    </body>

    <!-- The last step needed is to import the package -->
    <script type="module" src='https://unpkg.com/@roedoejet/readalong@latest/dist/read-along/read-along.esm.js'></script>
    <script nomodule src='https://unpkg.com/@roedoejet/readalong@latest/dist/read-along/read-along.js'></script>
</html>
"""


def save_minimal_index_html(
    output_path, tokenized_xml_basename, smil_basename, audio_basename
):
    with open(output_path, "w", encoding="utf-8") as fout:
        fout.write(
            MINIMAL_INDEX_HTML_TEMPLATE.format(
                tokenized_xml_basename, smil_basename, audio_basename
            )
        )


def unicode_normalize_xml(element):
    if element.text:
        element.text = normalize("NFD", unicode(element.text))
    for child in element.getchildren():
        unicode_normalize_xml(child)
        if child.tail:
            child.tail = normalize("NFD", unicode(child.tail))
